"""Loss functions.

Each loss function is intented to work with both single values as well as numpy vectors.

"""
import abc
import math
import typing

import numpy as np
from scipy import special

from creme import base
from creme import utils


__all__ = [
    'Absolute',
    'BinaryLoss',
    'BinaryFocalLoss',
    'Cauchy',
    'CrossEntropy',
    'Hinge',
    'EpsilonInsensitiveHinge',
    'Log',
    'MultiClassLoss',
    'Perceptron',
    'Poisson',
    'Quantile',
    'RegressionLoss',
    'Squared'
]


def clamp_proba(p):
    return max(min(p, 1 - 1e-15), 1e-15)


class Loss(abc.ABC):
    """Base class for all loss functions."""

    def __str__(self):
        return self.__class__.__name__

    @abc.abstractmethod
    def __call__(self, y_true, y_pred):
        """Returns the loss.

        Parameters:
            y_true: Ground truth(s).
            y_pred: Prediction(s).

        Returns:
            The losses.

        """

    @abc.abstractmethod
    def gradient(self, y_true, y_pred):
        """Returns the gradient with respect to y_pred.

        Parameters:
            y_true: Ground truth(s).
            y_pred: Prediction(s).

        Returns:
            The gradient(s).

        """

    @abc.abstractmethod
    def mean_func(self, y_pred):
        """Mean function.

        This is the inverse of the link function. Typically, a loss function takes as input the raw
        output of a model. In the case of classification, the raw output would be logits. The mean
        function can be used to convert the raw output into a value that makes sense to the user,
        such as a probability.

        Parameters:
            y_pred: Raw prediction(s).

        Returns:
            The adjusted prediction(s).

        References:
            1. [Wikipedia section on link and mean function](https://www.wikiwand.com/en/Generalized_linear_model#/Link_function)

        """


class BinaryLoss(Loss):
    """A loss appropriate for binary classification tasks."""

    def mean_func(self, y_pred):
        if isinstance(y_pred, np.ndarray):
            return 1. / (1. + np.exp(-y_pred))
        return utils.math.sigmoid(y_pred)


class MultiClassLoss(Loss):
    """A loss appropriate for multi-class classification tasks."""

    def mean_func(self, y_pred):
        if isinstance(y_pred, np.ndarray):
            return special.softmax(y_pred)
        return utils.math.softmax(y_pred)


class RegressionLoss(Loss):
    """A loss appropriate for regression tasks."""

    def mean_func(self, y_pred):
        return y_pred


class Absolute(RegressionLoss):
    """Absolute loss, also known as the mean absolute error or L1 loss.

    Mathematically, it is defined as

    $$L = |p_i - y_i|$$

    It's gradient w.r.t. to $p_i$ is

    $$\\frac{\\partial L}{\\partial p_i} = sgn(p_i - y_i)$$

    Example:

        >>> from creme import optim

        >>> loss = optim.losses.Absolute()
        >>> loss(-42, 42)
        84
        >>> loss.gradient(1, 2)
        1
        >>> loss.gradient(2, 1)
        -1

    """

    def __call__(self, y_true, y_pred):
        if isinstance(y_true, np.ndarray):
            return np.abs(y_pred - y_true)
        return abs(y_pred - y_true)

    def gradient(self, y_true, y_pred):

        if isinstance(y_true, np.ndarray):
            return np.where(y_pred > y_true, 1, -1)

        if y_pred > y_true:
            return 1
        return -1


class Cauchy(RegressionLoss):
    """Cauchy loss function.

    Parameters:
        C

    References:
        1. ["Effect of MAE" Kaggle discussion](https://www.kaggle.com/c/allstate-claims-severity/discussion/24520#140163)
        2. [Paris Madness Kaggle kernel](https://www.kaggle.com/raddar/paris-madness)

    """

    def __init__(self, C=80):
        self.C = C

    def __call__(self, y_true, y_pred):
        if isinstance(y_true, np.ndarray):
            return np.abs(y_pred - y_true)
        return abs(y_pred - y_true)

    def gradient(self, y_true, y_pred):
        diff = y_pred - y_true
        return diff / ((diff / self.C) ** 2 + 1)


class CrossEntropy(MultiClassLoss):
    """Cross entropy loss.

    This is a generalization of logistic loss to multiple classes.

    Parameters:
        class_weight: A dictionary that indicates what weight to associate with each class.

    Example:

        >>> from creme import optim

        >>> y_true = [0, 1, 2, 2]
        >>> y_pred = [
        ...     {0: 0.29450637, 1: 0.34216758, 2: 0.36332605},
        ...     {0: 0.21290077, 1: 0.32728332, 2: 0.45981591},
        ...     {0: 0.42860913, 1: 0.33380113, 2: 0.23758974},
        ...     {0: 0.44941979, 1: 0.32962558, 2: 0.22095463}
        ... ]

        >>> loss = optim.losses.CrossEntropy()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(loss(yt, yp))
        1.222454
        1.116929
        1.437209
        1.509797

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(loss.gradient(yt, yp))
        {0: -0.70549363, 1: 0.34216758, 2: 0.36332605}
        {0: 0.21290077, 1: -0.67271668, 2: 0.45981591}
        {0: 0.42860913, 1: 0.33380113, 2: -0.76241026}
        {0: 0.44941979, 1: 0.32962558, 2: -0.77904537}

    References:
        1. [What is Softmax regression and how is it related to Logistic regression?](https://github.com/rasbt/python-machine-learning-book/blob/master/faq/softmax_regression.md)

    """

    def __init__(self, class_weight: typing.Dict[base.typing.ClfTarget, float] = None):
        if class_weight is None:
            class_weight = {}
        self.class_weight = class_weight

    def __call__(self, y_true, y_pred):
        total = 0

        for label, proba in y_pred.items():
            if y_true == label:
                total += self.class_weight.get(label, 1.) * math.log(clamp_proba(proba))

        return -total

    def gradient(self, y_true, y_pred):
        return {
            label: (
                self.class_weight.get(label, 1.) *
                (clamp_proba(y_pred.get(label, 0.)) - (y_true == label))
            )
            for label in {*y_pred.keys(), y_true}
        }


class Hinge(BinaryLoss):
    """Computes the hinge loss.

    Mathematically, it is defined as

    $$L = max(0, 1 - p_i * y_i)$$

    It's gradient w.r.t. to $p_i$ is

    $$
    \\frac{\\partial L}{\\partial y_i} = \\left\{
    \\begin{array}{ll}
        \\ 0  &   p_iy_i \geqslant 1  \\\\
        \\ - y_i & p_iy_i < 1
    \\end{array}
    \\right.
    $$

    Parameters:
        threshold: Margin threshold. 1 yield the loss used in SVMs, whilst 0 is equivalent to
            the loss used in the Perceptron algorithm.

    Example:

        >>> from creme import optim

        >>> loss = optim.losses.Hinge(threshold=1)
        >>> loss(1, .2)
        0.8

        >>> loss.gradient(1, .2)
        -0.2

    """

    def __init__(self, threshold=1.):
        self.threshold = threshold

    def __call__(self, y_true, y_pred):
        y_true = y_true * 2 - 1  # [0, 1] -> [-1, 1]

        if isinstance(y_true, np.ndarray):
            return np.maximum(self.threshold - y_true * y_pred, 0)

        return max(self.threshold - y_true * y_pred, 0)

    def gradient(self, y_true, y_pred):
        y_true = y_true * 2 - 1  # [0, 1] -> [-1, 1]

        if isinstance(y_true, np.ndarray):
            return np.where(y_true * y_pred < self.threshold, -y_pred, 0)

        if y_true * y_pred < self.threshold:
            return -y_pred
        return 0


class EpsilonInsensitiveHinge(RegressionLoss):
    """Epsilon-insensitive hinge loss.

    Parameters:
        eps

    """

    def __init__(self, eps=.1):
        self.eps = eps

    def __call__(self, y_true, y_pred):
        y_true = y_true * 2 - 1  # [0, 1] -> [-1, 1]

        if isinstance(y_true, np.ndarray):
            return np.maximum(np.abs(y_pred - y_true) - self.eps, 0)

        return max(math.fabs(y_pred - y_true) - self.eps, 0)

    def gradient(self, y_true, y_pred):
        y_true = y_true * 2 - 1  # [0, 1] -> [-1, 1]

        if isinstance(y_true, np.ndarray):
            gradients = np.zeros_like(y_true)
            gradients[y_pred > y_true + self.eps] = 1
            gradients[y_pred + self.eps < y_true] = -1
            return gradients

        if y_pred > y_true + self.eps:
            return 1
        elif y_pred + self.eps < y_true:
            return -1
        return 0


class Log(BinaryLoss):
    """Logarithmic loss.

    This loss function expects each provided `y_pred` to be a logit. In other words if must be
    the raw output of a linear model or a neural network.

    Parameters:
        weight_pos
        weight_neg

    References:
        1. [Logit Wikipedia page](https://www.wikiwand.com/en/Logit>)

    """

    def __init__(self, weight_pos=1., weight_neg=1.):
        self.weight_pos = weight_pos
        self.weight_neg = weight_neg

    def __call__(self, y_true, y_pred):

        if isinstance(y_true, np.ndarray):
            weights = np.where(y_true == 0, self.weight_neg, self.weight_pos)
            z = y_pred * y_true
            return weights * np.log(1. + np.exp(-z))

        weight = self.weight_pos
        if y_true == 0:
            y_true = -1
            weight = self.weight_neg

        z = y_pred * y_true
        if z > 18.:
            return weight * math.exp(-z)
        if z < -18.:
            return weight * -z
        return weight * math.log(1. + math.exp(-z))

    def gradient(self, y_true, y_pred):

        if isinstance(y_true, np.ndarray):
            weights = np.where(y_true == 0, self.weight_neg, self.weight_pos)
            z = y_pred * y_true
            return weights * -y_true / (np.exp(z) + 1.)

        weight = self.weight_pos
        if y_true == 0:
            y_true = -1
            weight = self.weight_neg

        z = y_pred * y_true
        if z > 18.:
            return weight * math.exp(-z) * -y_true
        if z < -18.:
            return weight * -y_true
        return weight * -y_true / (math.exp(z) + 1.)


class Quantile(RegressionLoss):
    """Quantile loss.

    Parameters:
        alpha: Desired quantile to attain.

    Example:

        >>> from creme import optim

        >>> loss = optim.losses.Quantile(0.5)
        >>> loss(1, 3)
        1.0

        >>> loss.gradient(1, 3)
        0.5

        >>> loss.gradient(3, 1)
        -0.5

    References:
        1. [Wikipedia article on quantile regression](https://www.wikiwand.com/en/Quantile_regression)
        2. [Derivative from WolframAlpha](https://www.wolframalpha.com/input/?i=derivative+(y+-+p)+*+(alpha+-+Boole(y+-+p))+wrt+p)

    """

    def __init__(self, alpha=.5):
        self.alpha = alpha

    def __call__(self, y_true, y_pred):
        diff = y_pred - y_true
        return (self.alpha - (diff < 0)) * diff

    def gradient(self, y_true, y_pred):
        return (y_true < y_pred) - self.alpha


class Squared(RegressionLoss):
    """Squared loss, also known as the L2 loss.

    Mathematically, it is defined as

    $$L = (p_i - y_i) ^ 2$$

    It's gradient w.r.t. to $p_i$ is

    $$\\frac{\\partial L}{\\partial p_i} = 2 \times (p_i - y_i)$$

    One thing to note is that this convention is consistent with Vowpal Wabbit and PyTorch, but
    not with scikit-learn. Indeed scikit-learn divides the loss by 2, making the 2 dissapear in
    the gradient.

    Example:

        >>> from creme import optim

        >>> loss = optim.losses.Squared()
        >>> loss(-4, 5)
        81
        >>> loss.gradient(-4, 5)
        18
        >>> loss.gradient(5, -4)
        -18

    """

    def __call__(self, y_true, y_pred):
        return (y_pred - y_true) * (y_pred - y_true)

    def gradient(self, y_true, y_pred):
        return 2 * (y_pred - y_true)


class BinaryFocalLoss(BinaryLoss):
    """Binary focal loss.

    This implements the "star" algorithm from the appendix of the focal loss paper.

    Parameters:
        gamma
        beta

    Refenrences:
        1. [Lin, T.Y., Goyal, P., Girshick, R., He, K. and Dollár, P., 2017. Focal loss for dense object detection. In Proceedings of the IEEE international conference on computer vision (pp. 2980-2988)](https://arxiv.org/pdf/1708.02002.pdf)

    """

    def __init__(self, gamma=2, beta=1):
        self.gamma = gamma
        self.beta = beta

    def __call__(self, y_true, y_pred):

        y_true = y_true * 2 - 1  # [0, 1] -> [-1, 1]

        xt = y_true * y_pred

        if isinstance(y_true, np.ndarray):
            pt = 1. / (1 + np.exp(-(self.gamma * xt + self.beta)))
            return -np.log(pt) / self.gamma

        pt = utils.math.sigmoid(self.gamma * xt + self.beta)
        return -math.log(pt) / self.gamma

    def gradient(self, y_true, y_pred):

        y_true = y_true * 2 - 1  # [0, 1] -> [-1, 1]

        xt = y_true * y_pred

        if isinstance(y_true, np.ndarray):
            pt = 1. / (1 + np.exp(-(self.gamma * xt + self.beta)))
            return y_true * (pt - 1)

        pt = utils.math.sigmoid(self.gamma * xt + self.beta)
        return y_true * (pt - 1)


class Poisson(RegressionLoss):
    """Poisson loss.

    The Poisson loss is usually more suited for regression with count data than the squared loss.

    Mathematically, it is defined as

    $$L = exp(p_i) - y_i \\times p_i$$

    It's gradient w.r.t. to $p_i$ is

    $$\\frac{\\partial L}{\\partial p_i} = exp(p_i) - y_i$$

    """

    def __call__(self, y_true, y_pred):
        if isinstance(y_pred, np.ndarray):
            return np.exp(y_pred) - y_true * y_pred
        return math.exp(y_pred) - y_true * y_pred

    def gradient(self, y_true, y_pred):
        if isinstance(y_pred, np.ndarray):
            return np.exp(y_pred) - y_true
        return math.exp(y_pred) - y_true

    def mean_func(self, y_pred):
        if isinstance(y_pred, np.ndarray):
            return np.exp(y_pred)
        return math.exp(y_pred)


class Perceptron(Hinge):
    """Perceptron loss.

    The Perceptron loss is the loss used in the Perceptron algorithm. Using this loss in a logistic
    regression yields the Perceptron algorithm.

    Mathematically, it is defined as

    $$L = exp(p_i) - y_i \\times p_i$$

    It's gradient w.r.t. to $p_i$ is

    $$\\frac{\\partial L}{\\partial p_i} = exp(p_i) - y_i$$

    References:
        1. [Wikipedia page on the Perceptron algorithm](https://www.wikiwand.com/en/Perceptron)

    """

    def __init__(self):
        super().__init__(threshold=0.)



"""Dummy estimators.

This module is here for testing purposes, as well as providing baseline performances.

"""
import collections

from creme import base
from creme import stats


__all__ = ['NoChangeClassifier', 'PriorClassifier', 'StatisticRegressor']


class NoChangeClassifier(base.MultiClassifier):
    """Dummy classifier which returns the last class seen.

    The predict_one method will output the last class seen whilst predict_proba_one will
    return 1 for the last class seen and 0 for the others.

    Attributes:
        last_class: The last class seen.
        classes: The set of classes seen.

    Example:

        Taken from example 2.1 from [this page](https://www.cms.waikato.ac.nz/~abifet/book/chapter_2.html).

        >>> import pprint
        >>> from creme import dummy

        >>> sentences = [
        ...     ('glad happy glad', '+'),
        ...     ('glad glad joyful', '+'),
        ...     ('glad pleasant', '+'),
        ...     ('miserable sad glad', '−')
        ... ]

        >>> model = dummy.NoChangeClassifier()

        >>> for sentence, label in sentences:
        ...     model = model.fit_one(sentence, label)

        >>> new_sentence = 'glad sad miserable pleasant glad'
        >>> model.predict_one(new_sentence)
        '−'

        >>> pprint.pprint(model.predict_proba_one(new_sentence))
        {'+': 0, '−': 1}

    """

    def __init__(self):
        self.last_class = None
        self.classes = set()

    def fit_one(self, x, y):
        self.last_class = y
        self.classes.add(y)
        return self

    def predict_one(self, x):
        return self.last_class

    def predict_proba_one(self, x):
        probas = {c: 0 for c in self.classes}
        probas[self.last_class] = 1
        return probas


class PriorClassifier(base.MultiClassifier):
    """Dummy classifier which uses the prior distribution.

    The `predict_one` method will output the most common class whilst `predict_proba_one` will
    return the normalized class counts.

    Attributes:
        counts (collections.Counter): Class counts.
        n (int): Total number of seen instances.

    Example:

        Taken from example 2.1 from [this page](https://www.cms.waikato.ac.nz/~abifet/book/chapter_2.html)

        >>> from creme import dummy

        >>> sentences = [
        ...     ('glad happy glad', '+'),
        ...     ('glad glad joyful', '+'),
        ...     ('glad pleasant', '+'),
        ...     ('miserable sad glad', '−')
        ... ]

        >>> model = dummy.PriorClassifier()

        >>> for sentence, label in sentences:
        ...     model = model.fit_one(sentence, label)

        >>> new_sentence = 'glad sad miserable pleasant glad'
        >>> model.predict_one(new_sentence)
        '+'
        >>> model.predict_proba_one(new_sentence)
        {'+': 0.75, '−': 0.25}

    """

    def __init__(self):
        self.counts = collections.Counter()
        self.n = 0

    def fit_one(self, x, y):
        self.counts.update([y])
        self.n += 1
        return self

    def predict_proba_one(self, x):
        return {label: count / self.n for label, count in self.counts.items()}


class StatisticRegressor(base.Regressor):
    """Dummy regressor that uses a univariate statistic to make predictions.

    Parameters:
        statistic

    Example:

        >>> from pprint import pprint
        >>> from creme import dummy
        >>> from creme import stats

        >>> sentences = [
        ...     ('glad happy glad', 3),
        ...     ('glad glad joyful', 3),
        ...     ('glad pleasant', 2),
        ...     ('miserable sad glad', -3)
        ... ]

        >>> model = dummy.StatisticRegressor(stats.Mean())

        >>> for sentence, score in sentences:
        ...     model = model.fit_one(sentence, score)

        >>> new_sentence = 'glad sad miserable pleasant glad'
        >>> model.predict_one(new_sentence)
        1.25

    """

    def __init__(self, statistic: stats.Univariate):
        self.statistic = statistic

    def fit_one(self, x, y):
        self.statistic.update(y)
        return self

    def predict_one(self, x):
        return self.statistic.get()



"""Feature preprocessing."""
from .feature_hasher import FeatureHasher
from .kernel_approx import RBFSampler
from .one_hot import OneHotEncoder
from .poly import PolynomialExtender
from .scale import Binarizer
from .scale import MaxAbsScaler
from .scale import MinMaxScaler
from .scale import Normalizer
from .scale import RobustScaler
from .scale import StandardScaler


__all__ = [
    'Binarizer',
    'FeatureHasher',
    'MaxAbsScaler',
    'MinMaxScaler',
    'Normalizer',
    'OneHotEncoder',
    'PolynomialExtender',
    'RBFSampler',
    'RobustScaler',
    'StandardScaler'
]



import collections
import math

from . import base


__all__ = ['ComplementNB']


class ComplementNB(base.BaseNB):
    """Naive Bayes classifier for multinomial models.

    The input vector has to contain positive values, such as counts or TF-IDF values.

    This class inherits `predict_proba_one` from `naive_bayes.BaseNB` which itself inherits
    `predict_one` from `base.MultiClassifier`.

    Parameters:
        alpha: Additive (Laplace/Lidstone) smoothing parameter (use 0 for no smoothing).

    Attributes:
        class_dist (proba.Multinomial): Class prior probability distribution.
        feature_counts (collections.defaultdict): Total frequencies per feature and class.
        class_totals (collections.Counter): Total frequencies per class.

    Example:

        ::

            >>> from creme import feature_extraction
            >>> from creme import naive_bayes

            >>> sentences = [
            ...     ('food food meat brain', 'health'),
            ...     ('food meat ' + 'kitchen ' * 9 + 'job' * 5, 'butcher'),
            ...     ('food food meat job', 'health')
            ... ]

            >>> model = feature_extraction.BagOfWords() | ('nb', naive_bayes.ComplementNB)

            >>> for sentence, label in sentences:
            ...     model = model.fit_one(sentence, label)

            >>> model['nb'].p_class('health') == 2 / 3
            True
            >>> model['nb'].p_class('butcher') == 1 / 3
            True

            >>> model.predict_proba_one('food job meat')
            {'health': 0.779191, 'butcher': 0.220808}

    References:
        1. [Rennie, J.D., Shih, L., Teevan, J. and Karger, D.R., 2003. Tackling the poor assumptions of naive bayes text classifiers. In Proceedings of the 20th international conference on machine learning (ICML-03) (pp. 616-623)](https://people.csail.mit.edu/jrennie/papers/icml03-nb.pdf)
        2. [StackExchange discussion](https://stats.stackexchange.com/questions/126009/complement-naive-bayes)

    """

    def __init__(self, alpha=1.):
        self.alpha = alpha
        self.class_counts = collections.Counter()
        self.feature_counts = collections.defaultdict(collections.Counter)
        self.feature_totals = collections.Counter()
        self.class_totals = collections.Counter()

    def fit_one(self, x, y):
        self.class_counts.update((y,))

        for f, frequency in x.items():
            self.feature_counts[f].update({y: frequency})
            self.feature_totals.update({f: frequency})
            self.class_totals.update({y: frequency})

        return self

    def p_class(self, c):
        return self.class_counts[c] / sum(self.class_counts.values())

    def joint_log_likelihood(self, x):
        return {
            c: sum(
                (
                    frequency * -math.log(
                        (self.feature_totals[f] - self.feature_counts.get(f, {}).get(c, 0.) + 1) /
                        (self.class_totals[c] + 1 * len(self.feature_counts))
                    )
                    for f, frequency in x.items()
                )
            )
            for c in self.class_counts
        }



import math
import numbers
import collections

from .. import utils

from . import base


__all__ = ['AdaBound']


class AdaBound(base.Optimizer):
    """AdaBound optimizer.

    Parameters:
        lr: The learning rate.
        beta_1
        beta_2
        eps
        gamma
        final_lr

    Attributes:
        m (collections.defaultdict)
        s (collections.defaultdict)

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.AdaBound()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.879004

    References:
        1. [Luo, L., Xiong, Y., Liu, Y. and Sun, X., 2019. Adaptive gradient methods with dynamic bound of learning rate. arXiv preprint arXiv:1902.09843](https://arxiv.org/abs/1902.09843)

    """

    def __init__(self, lr=1e-3, beta_1=0.9, beta_2=0.999, eps=1e-8, gamma=1e-3, final_lr=0.1):

        if not isinstance(lr, numbers.Number):
            raise ValueError(
                f'lr in AdaBound should be numeric but got {type(lr)}')

        if not isinstance(final_lr, numbers.Number):
            raise ValueError(
                f'final_lr in AdaBound should be numeric but got {type(final_lr)}')

        super().__init__(lr)
        self.base_lr = lr
        self.final_lr = final_lr
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.eps = eps
        self.gamma = gamma
        self.m = collections.defaultdict(float)
        self.v = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        bias_1 = 1 - self.beta_1 ** (self.n_iterations + 1)
        bias_2 = 1 - self.beta_2 ** (self.n_iterations + 1)

        step_size = self.learning_rate * math.sqrt(bias_2) / bias_1
        self.final_lr *= self.learning_rate / self.base_lr

        lower_bound = self.final_lr * (1 - 1 / (self.gamma * (self.n_iterations + 1) + 1))
        upper_bound = self.final_lr * (1 + 1 / (self.gamma * (self.n_iterations + 1)))

        for i, gi in g.items():
            self.m[i] = self.beta_1 * self.m[i] + (1 - self.beta_1) * gi
            self.v[i] = self.beta_2 * self.v[i] + (1 - self.beta_2) * gi ** 2

            step_size_bound = step_size / (math.sqrt(self.v[i]) + self.eps)

            w[i] -= utils.math.clamp(step_size_bound, lower_bound, upper_bound) * self.m[i]

        return w



import collections
import functools
import math

from creme import proba

from . import base


__all__ = ['GaussianNB']


class GaussianNB(base.BaseNB):
    """Gaussian Naive Bayes.

    A Gaussian distribution $G_{cf}$ is maintained for each class $c$ and each feature $f$. Each
    Gaussian is updated using the amount associated with each feature; the details can be be found
    in `proba.Gaussian`. The joint log-likelihood is then obtained by summing the log probabilities
    of each feature associated with each class.

    This class inherits `predict_proba_one` from `naive_bayes.BaseNB` which itself inherits
    `predict_one` from `base.MultiClassifier`.

    Example:

        >>> from creme import naive_bayes
        >>> from creme import stream
        >>> import numpy as np

        >>> X = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
        >>> Y = np.array([1, 1, 1, 2, 2, 2])

        >>> model = naive_bayes.GaussianNB()

        >>> for x, y in stream.iter_array(X, Y):
        ...     _ = model.fit_one(x, y)

        >>> model.predict_one({0: -0.8, 1: -1})
        1

    """

    def __init__(self):
        self.class_counts = collections.Counter()
        defaultdict = collections.defaultdict
        self.gaussians = defaultdict(functools.partial(defaultdict, proba.Gaussian))

    def fit_one(self, x, y):

        self.class_counts.update((y,))

        for i, xi in x.items():
            self.gaussians[y][i].update(xi)

        return self

    def p_class(self, c):
        return self.class_counts[c] / sum(self.class_counts.values())

    def joint_log_likelihood(self, x):
        return {
            c: math.log(self.p_class(c)) + sum(
                math.log(10e-10 + gaussians[i].pdf(xi))
                for i, xi in x.items()
            )
            for c, gaussians in self.gaussians.items()
        }



import math

from creme import utils

from . import base


class Quantile(base.Univariate):
    """Running quantile.

    Uses the P-square algorithm. The code is inspired by LiveStat's implementation [2].

    Parameters:
        q: Determines which quantile to compute, must be comprised between 0 and 1.

    Example:

        >>> from creme import stats
        >>> import numpy as np

        >>> np.random.seed(42 * 1337)
        >>> mu, sigma = 0, 1
        >>> s = np.random.normal(mu, sigma, 500)

        >>> median = stats.Quantile(0.5)
        >>> for x in s:
        ...    _ = median.update(x)

        >>> print(f'The estimated value of the 50th (median) quantile is {median.get():.4f}')
        The estimated value of the 50th (median) quantile is -0.0275
        >>> print(f'The real value of the 50th (median) quantile is {np.median(s):.4f}')
        The real value of the 50th (median) quantile is -0.0135

        >>> percentile_17 = stats.Quantile(0.17)
        >>> for x in s:
        ...    _ = percentile_17.update(x)

        >>> print(f'The estimated value of the 17th quantile is {percentile_17.get():.4f}')
        The estimated value of the 17th quantile is -0.8652
        >>> print(f'The real value of the 17th quantile is {np.percentile(s,17):.4f}')
        The real value of the 17th quantile is -0.9072

    References:
        1. [The P2 Algorithm for Dynamic Univariateal Computing Calculation of Quantiles and Editor Histograms Without Storing Observations](https://www.cse.wustl.edu/~jain/papers/ftp/psqr.pdf)

    """

    def __init__(self, q=.5):

        if not 0 < q < 1:
            raise ValueError('q is not comprised between 0 and 1')
        self.q = q

        self.desired_marker_position = [
            0,
            self.q / 2,
            self.q,
            (1 + self.q) / 2,
            1
        ]
        self.marker_position = [
            1,
            1 + 2 * self.q,
            1 + 4 * self.q,
            3 + 2 * self.q,
            5
        ]
        self.position = list(range(1, 6))
        self.heights = []
        self.heights_sorted = False

    def _find_k(self, x):

        if x < self.heights[0]:
            self.heights[0] = x
            k = 1

        else:
            for i in range(1, 5):
                if self.heights[i - 1] <= x and x < self.heights[i]:
                    k = i
                    break
            else:
                k = 4
                if self.heights[-1] < x:
                    self.heights[-1] = x
        return k

    @classmethod
    def _compute_P2(cls, qp1, q, qm1, d, np1, n, nm1):

        d = float(d)
        n = float(n)
        np1 = float(np1)
        nm1 = float(nm1)

        outer = d / (np1 - nm1)
        inner_left = (n - nm1 + d) * (qp1 - q) / (np1 - n)
        inner_right = (np1 - n - d) * (q - qm1) / (n - nm1)

        return q + outer * (inner_left + inner_right)

    def _adjust(self):

        for i in range(1, 4):
            n = self.position[i]
            q = self.heights[i]

            d = self.marker_position[i] - n

            if (d >= 1 and self.position[i + 1] - n > 1) or (d <= -1 and self.position[i - 1] - n < -1):
                d = int(math.copysign(1, d))

                qp1 = self.heights[i + 1]
                qm1 = self.heights[i - 1]
                np1 = self.position[i + 1]
                nm1 = self.position[i - 1]

                qn = self._compute_P2(qp1, q, qm1, d, np1, n, nm1)

                if qm1 < qn and qn < qp1:
                    self.heights[i] = qn
                else:
                    self.heights[i] = q + d * \
                        (self.heights[i + d] - q) / (self.position[i + d] - n)

                self.position[i] = n + d

        return self

    def update(self, x):

        # Initialisation
        if len(self.heights) != 5:
            self.heights.append(x)

        else:
            if not self.heights_sorted:
                self.heights.sort()
                self.heights_sorted = True

            # Find cell k such that qk < Xj <= qk+i and adjust extreme values (q1 and q) if necessary
            k = self._find_k(x)

            # Increment all positions greater than k
            self.position = [j if i < k else j +
                             1 for i, j in enumerate(self.position)]
            self.marker_position = [
                x + y
                for x, y in zip(self.marker_position, self.desired_marker_position)
            ]

            # Adjust heights of markers 2-4 if necessary
            self._adjust()

        return self

    def get(self):
        if self.heights_sorted:
            return self.heights[2]

        if self.heights:
            self.heights.sort()
            length = len(self.heights)
            return self.heights[int(min(max(length - 1, 0), length * self.q))]

        return 0


class RollingQuantile(base.RollingUnivariate, utils.SortedWindow):
    """Running quantile over a window.

    Parameters:
        q: Determines which quantile to compute, must be comprised between 0 and 1.
        window_size: Size of the window.

    Example:

        >>> from creme import stats

        >>> rolling_quantile = stats.RollingQuantile(
        ...     q=.5,
        ...     window_size=100,
        ... )

        >>> for i in range(0, 1001):
        ...     rolling_quantile = rolling_quantile.update(i)
        ...     if i % 100 == 0:
        ...         print(rolling_quantile.get())
        0
        50
        150
        250
        350
        450
        550
        650
        750
        850
        950

    References:
        1. [Left sorted](https://stackoverflow.com/questions/8024571/insert-an-item-into-sorted-list-in-python)

    """

    def __init__(self, q, window_size):
        super().__init__(size=window_size)
        self.q = q
        self.idx = int(round(self.q * self.size + .5)) - 1

    @property
    def window_size(self):
        return self.size

    def update(self, x):
        self.append(x)
        return self

    def get(self):
        if len(self) < self.size:
            idx = int(round(self.q * len(self) + .5)) - 1
            return self[idx]
        return self[self.idx]



import abc
import collections
import numbers
import typing

from creme import base
from creme import utils
from creme import stats
from creme.reco.base import Recommender


__all__ = [
    'BinaryMetric',
    'ClassificationMetric',
    'Metric',
    'Metrics',
    'MultiClassMetric',
    'MultiOutputClassificationMetric',
    'MultiOutputRegressionMetric',
    'RegressionMetric',
    'WrapperMetric'
]


class Metric(abc.ABC):

    # Define the format specification used for string representation.
    _fmt = ',.6f'  # Use commas to separate big numbers and show 6 decimals

    @abc.abstractmethod
    def update(self, y_true, y_pred) -> 'Metric':
        """Update the metric."""

    @abc.abstractmethod
    def get(self) -> float:
        """Return the current value of the metric."""

    @abc.abstractproperty
    def bigger_is_better(self) -> bool:
        """Indicates if a high value is better than a low one or not."""

    @abc.abstractmethod
    def works_with(self, model: base.Estimator) -> bool:
        """Indicates whether or not a metric can work with a given model."""

    def __repr__(self):
        """Returns the class name along with the current value of the metric."""
        return f'{self.__class__.__name__}: {self.get():{self._fmt}}'.rstrip('0')


class ClassificationMetric(Metric):

    @abc.abstractproperty
    def requires_labels(self) -> bool:
        """Indicates if labels are required, rather than probabilities."""

    @staticmethod
    def _clamp_proba(p):
        return utils.math.clamp(p, minimum=1e-15, maximum=1 - 1e-15)

    def __add__(self, other) -> 'Metrics':
        if not isinstance(other, ClassificationMetric):
            raise ValueError(f'{self.__class__.__name__} and {other.__class__.__name__} metrics '
                             'are not compatible')
        return Metrics([self, other])


class BinaryMetric(ClassificationMetric):

    @abc.abstractmethod
    def update(
        self,
        y_true: bool,
        y_pred: typing.Union[bool, float, typing.Dict[bool, float]],
        sample_weight: numbers.Number
    ) -> 'BinaryMetric':
        """Update the metric."""

    @abc.abstractmethod
    def revert(
        self,
        y_true: bool,
        y_pred: typing.Union[bool, float, typing.Dict[bool, float]],
        sample_weight: numbers.Number
    ) -> 'BinaryMetric':
        """Revert the metric."""

    def works_with(self, model) -> bool:
        return isinstance(utils.estimator_checks.guess_model(model), base.Classifier)


class MultiClassMetric(ClassificationMetric):

    @abc.abstractmethod
    def update(
        self,
        y_true: base.typing.ClfTarget,
        y_pred: typing.Union[base.typing.ClfTarget, typing.Dict[base.typing.ClfTarget, float]],
        sample_weight: numbers.Number
    ) -> 'MultiClassMetric':
        """Update the metric."""

    @abc.abstractmethod
    def revert(
        self,
        y_true: bool,
        y_pred: typing.Union[base.typing.ClfTarget, typing.Dict[base.typing.ClfTarget, float]],
        sample_weight: numbers.Number
    ) -> 'MultiClassMetric':
        """Revert the metric."""

    def works_with(self, model) -> bool:
        return isinstance(utils.estimator_checks.guess_model(model), base.Classifier)


class RegressionMetric(Metric):

    @abc.abstractmethod
    def update(
        self,
        y_true: numbers.Number,
        y_pred: numbers.Number,
        sample_weight: numbers.Number
    ) -> 'RegressionMetric':
        """Update the metric."""

    @abc.abstractmethod
    def revert(
        self,
        y_true: numbers.Number,
        y_pred: numbers.Number,
        sample_weight: numbers.Number
    ) -> 'RegressionMetric':
        """Revert the metric."""

    @property
    def bigger_is_better(self):
        return False

    def works_with(self, model) -> bool:
        return isinstance(utils.estimator_checks.guess_model(model), base.Regressor)

    def __add__(self, other) -> 'Metrics':
        if not isinstance(other, RegressionMetric):
            raise ValueError(f'{self.__class__.__name__} and {other.__class__.__name__} metrics '
                             'are not compatible')
        return Metrics([self, other])


class MultiOutputClassificationMetric(Metric):

    def update(
        self,
        y_true: typing.Dict[typing.Union[str, int], base.typing.ClfTarget],
        y_pred: typing.Union[
            typing.Dict[typing.Union[str, int], base.typing.ClfTarget],
            typing.Dict[typing.Union[str, int], typing.Dict[base.typing.ClfTarget, float]]
        ],
        sample_weight: numbers.Number
    ) -> 'MultiOutputClassificationMetric':
        """Update the metric."""

    def revert(
        self,
        y_true: typing.Dict[typing.Union[str, int], base.typing.ClfTarget],
        y_pred: typing.Union[
            typing.Dict[typing.Union[str, int], base.typing.ClfTarget],
            typing.Dict[typing.Union[str, int], typing.Dict[base.typing.ClfTarget, float]]
        ],
        sample_weight: numbers.Number
    ) -> 'MultiOutputClassificationMetric':
        """Revert the metric."""

    def works_with(self, model) -> bool:
        return isinstance(utils.estimator_checks.guess_model(model), base.MultiOutputClassifier)


class MultiOutputRegressionMetric(Metric):

    def update(
        self,
        y_true: typing.Dict[typing.Union[str, int], typing.Union[float, int]],
        y_pred: typing.Dict[typing.Union[str, int], typing.Union[float, int]],
        sample_weight: numbers.Number
    ) -> 'MultiOutputRegressionMetric':
        """Update the metric."""

    def revert(
        self,
        y_true: typing.Dict[typing.Union[str, int], typing.Union[float, int]],
        y_pred: typing.Dict[typing.Union[str, int], typing.Union[float, int]],
        sample_weight: numbers.Number
    ) -> 'MultiOutputRegressionMetric':
        """Revert the metric."""

    def works_with(self, model) -> bool:
        return isinstance(utils.estimator_checks.guess_model(model), base.MultiOutputRegressor)


class Metrics(Metric, collections.UserList):
    """A container class for handling multiple metrics at once."""

    def __init__(self, metrics, str_sep=', '):
        super().__init__(metrics)
        self.str_sep = str_sep

    def update(self, y_true, y_pred, sample_weight=1.):

        # If the metrics are classification metrics, then we have to handle the case where some
        # of the metrics require labels, whilst others need to be fed probabilities
        if hasattr(self, 'requires_labels') and not self.requires_labels:
            for m in self:
                if m.requires_labels:
                    m.update(y_true, max(y_pred, key=y_pred.get))
                else:
                    m.update(y_true, y_pred)
            return self

        for m in self:
            m.update(y_true, y_pred)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):

        # If the metrics are classification metrics, then we have to handle the case where some
        # of the metrics require labels, whilst others need to be fed probabilities
        if hasattr(self, 'requires_labels') and not self.requires_labels:
            for m in self:
                if m.requires_labels:
                    m.revert(y_true, max(y_pred, key=y_pred.get), sample_weight)
                else:
                    m.revert(y_true, y_pred, sample_weight)
            return self

        for m in self:
            m.revert(y_true, y_pred, sample_weight)
        return self

    def get(self):
        return [m.get() for m in self]

    def works_with(self, model) -> bool:
        return all(m.works_with(utils.estimator_checks.guess_model(model)) for m in self)

    @property
    def bigger_is_better(self):
        raise NotImplementedError

    @property
    def requires_labels(self):
        return all(m.requires_labels for m in self)

    def __repr__(self):
        return self.str_sep.join((str(m) for m in self))

    def __add__(self, other):
        try:
            other + self[0]  # Will raise a ValueError if incompatible
        except IndexError:
            pass
        self.append(other)
        return self


class WrapperMetric(Metric):

    @property
    def _fmt(self):
        return self.metric._fmt

    @abc.abstractproperty
    def metric(self):
        """Gives access to the wrapped metric."""

    def get(self):
        return self.metric.get()

    @property
    def bigger_is_better(self):
        return self.metric.bigger_is_better

    def works_with(self, model):
        return self.metric.works_with(model)

    @property
    def requires_labels(self):
        return self.metric.requires_labels

    @property
    def __metaclass__(self):
        return self.metric.__class__

    def __repr__(self):
        return str(self.metric)


class MeanMetric(abc.ABC):
    """Many metrics are just running averages. This is a utility class that avoids repeating
    tedious stuff throughout the module for such metrics.

    """

    def __init__(self):
        self._mean = stats.Mean()

    @abc.abstractmethod
    def _eval(self, y_true, y_pred):
        pass

    def update(self, y_true, y_pred, sample_weight=1.):
        self._mean.update(x=self._eval(y_true, y_pred), w=sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self._mean.revert(x=self._eval(y_true, y_pred), w=sample_weight)
        return self

    def get(self):
        return self._mean.get()



from . import mse


__all__ = ['RMSE']


class RMSE(mse.MSE):
    """Root mean squared error.

    Example:

        >>> from creme import metrics

        >>> y_true = [3, -0.5, 2, 7]
        >>> y_pred = [2.5, 0.0, 2, 8]

        >>> metric = metrics.RMSE()
        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp).get())
        0.5
        0.5
        0.408248
        0.612372

        >>> metric
        RMSE: 0.612372

    """

    def get(self):
        return super().get() ** .5



import copy
import functools
import math
import operator
import typing

from creme import base
from creme import metrics


__all__ = ['SuccessiveHalvingClassifier', 'SuccessiveHalvingRegressor']


class SuccessiveHalving:

    def __init__(self, models: typing.List[base.Predictor], metric: metrics.Metric, budget: int,
                 eta=2, verbose=False, **print_kwargs):

        # Check that the model and the metric are in accordance
        for model in models:
            if not metric.works_with(model):
                raise ValueError(f"{metric.__class__.__name__} metric can't be used to evaluate a " +
                                f'{model.__class__.__name__}')

        self.models = models
        self.metric = metric
        self.budget = budget
        self.eta = eta
        self.verbose = verbose
        self.print_kwargs = print_kwargs

        self._n = len(models)
        self._metrics = [copy.deepcopy(metric) for _ in range(self._n)]
        self._s = self._n
        self._n_rungs = 0
        self._rankings = list(range(self._n))
        self._r = math.floor(budget / (self._s * math.ceil(math.log(self._n, eta))))
        self._budget_used = 0
        self._n_iterations = 0
        self._best_model_idx = 0

        if isinstance(model, base.Classifier) and not metric.requires_labels:
            self._pred_func = lambda model: model.predict_proba_one
        else:
            self._pred_func = lambda model: model.predict_one

    @property
    def best_model(self):
        """The current best model."""
        return self.models[self._best_model_idx]

    def fit_one(self, x, y):

        for i in self._rankings[:self._s]:
            model = self.models[i]
            metric = self._metrics[i]
            y_pred = self._pred_func(model)(x)
            metric.update(y_true=y, y_pred=y_pred)
            model.fit_one(x, y)

            # Check for a new best model
            if i != self._best_model_idx:
                op = operator.gt if self.metric.bigger_is_better else operator.lt
                if op(self._metrics[i].get(), self._metrics[self._best_model_idx].get()):
                    self._best_model_idx = i

        self._n_iterations += 1

        if self._s > 1 and self._n_iterations == self._r:

            self._n_rungs += 1
            self._budget_used += self._s * self._r

            # Update the rankings of the current models based on their respective metric values
            self._rankings[:self._s] = sorted(
                self._rankings[:self._s],
                key=lambda i: self._metrics[i].get(),
                reverse=self.metric.bigger_is_better
            )

            # Determine how many models to keep for the current rung
            cutoff = math.ceil(self._s / self.eta)

            if self.verbose:
                print(
                    '\t'.join((
                        f'[{self._n_rungs}]',
                        f'{self._s - cutoff} removed',
                        f'{cutoff} left',
                        f'{self._r} iterations',
                        f'budget used: {self._budget_used}',
                        f'budget left: {self.budget - self._budget_used}',
                        f'best {self._metrics[self._rankings[0]]}',
                    )),
                    **self.print_kwargs
                )

            # Determine where the next rung is located
            self._s = cutoff
            self._r = math.floor(self.budget / (self._s * math.ceil(math.log(self._n, self.eta))))

        return self


class SuccessiveHalvingRegressor(SuccessiveHalving, base.Regressor):
    """Successive halving algorithm for regression.

    Successive halving is a method for performing model selection without having to train each
    model on all the dataset. At certain points in time (called "rungs"), the worst performing will
    be discarded and the best ones will keep competing between each other. The rung values are
    designed so that at most `budget` model updates will be performed in total.

    If you have `k` combinations of hyperparameters and that your dataset contains `n`
    observations, then the maximal budget you can allocate is:

    $$\\frac{2kn}{eta}$$

    It is recommended that you check this beforehand. This bound can't be checked by the function
    because the size of the dataset is not known. In fact it is potentially infinite, in which case
    the algorithm will terminate once all the budget has been spent.

    If you have a budget of `B`, and that your dataset contains `n` observations, then the
    number of hyperparameter combinations that will spend all the budget and go through all the
    data is:

    $$\\ceil(\\floor(\\frac{B}{2n}) \\times eta)$$

    Parameters:
        models: The models to compare.
        X_y: A stream of (features, target) tuples.
        metric: Metric used for comparing models with.
        budget: Total number of model updates you wish to allocate.
        eta: Rate of elimination. At every rung, `math.ceil(k / eta)` models are kept, where
            `k` is the number of models that have reached the rung. A higher `eta` value will
            focus on less models but will allocate more iterations to the best models.
        verbose: Whether to display progress or not.
        print_kwargs: Extra keyword arguments are passed to the `print` function. For instance,
            this allows providing a `file` argument, which indicates where to output progress.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import preprocessing
        >>> from creme import model_selection
        >>> from creme import optim

        As an example, let's use successive halving to tune the learning rate of a linear
        regression. We'll first define the model.

        >>> from creme import linear_model
        >>> from creme import preprocessing

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LinearRegression(intercept_lr=.1)
        ... )

        Let's now define a grid of parameters which we would like to compare. We'll try
        different optimizers with various learning rates.

        >>> param_grid = model_selection.expand_param_grid({
        ...     'LinearRegression': {
        ...         'optimizer': [
        ...             (optim.SGD, {'lr': [.1, .01, .005]}),
        ...             (optim.Adam, {'beta_1': [.01, .001], 'lr': [.1, .01, .001]}),
        ...             (optim.Adam, {'beta_1': [.1], 'lr': [.001]}),
        ...         ]
        ...     }
        ... })

        Let's now instantiate multiple copies of the model with each parameter configuration.

        >>> models = [model._set_params(params) for params in param_grid]

        We can check how many models this created.

        >>> len(models)
        10

        We can now pass these models to a `SuccessiveHalvingRegressor`. We also need to pick a
        metric to compare the models, and a budget which indicates how many iterations to run
        before picking the best model and discarding the rest.

        >>> sh = model_selection.SuccessiveHalvingRegressor(
        ...     models=models,
        ...     metric=metrics.MAE(),
        ...     budget=2000,
        ...     eta=2,
        ...     verbose=True
        ... )

        A `SuccessiveHalvingRegressor` is also a regressor with a `fit_one` and a `predict_one`
        method, therefore we can use `creme.model_selection.progressive_val_score` to evaluate it.

        >>> model_selection.progressive_val_score(
        ...     X_y=datasets.TrumpApproval(),
        ...     model=sh,
        ...     metric=metrics.MAE()
        ... )
        [1]	5 removed	5 left	50 iterations	budget used: 500	budget left: 1500	best MAE: 4.540491
        [2]	2 removed	3 left	100 iterations	budget used: 1000	budget left: 1000	best MAE: 2.458765
        [3]	1 removed	2 left	166 iterations	budget used: 1498	budget left: 502	best MAE: 1.583751
        [4]	1 removed	1 left	250 iterations	budget used: 1998	budget left: 2	best MAE: 1.147296
        MAE: 0.488387

        We can now view the best model.

        >>> sh.best_model
        Pipeline (
          StandardScaler (),
          LinearRegression (
            optimizer=Adam (
              lr=Constant (
                learning_rate=0.1
              )
              beta_1=0.01
              beta_2=0.999
              eps=1e-08
            )
            loss=Squared ()
            l2=0.
            intercept=39.93843
            intercept_lr=Constant (
              learning_rate=0.1
            )
            clip_gradient=1e+12
            initializer=Zeros ()
          )
        )

    References:
        1. [Jamieson, K. and Talwalkar, A., 2016, May. Non-stochastic best arm identification and hyperparameter optimization. In Artificial Intelligence and Statistics (pp. 240-248).](http://proceedings.mlr.press/v51/jamieson16.pdf)
        2. [Li, L., Jamieson, K., Rostamizadeh, A., Gonina, E., Hardt, M., Recht, B. and Talwalkar, A., 2018. Massively parallel hyperparameter tuning. arXiv preprint arXiv:1810.05934.](https://arxiv.org/pdf/1810.05934.pdf)
        3. [Li, L., Jamieson, K., DeSalvo, G., Rostamizadeh, A. and Talwalkar, A., 2017. Hyperband: A novel bandit-based approach to hyperparameter optimization. The Journal of Machine Learning Research, 18(1), pp.6765-6816.](https://arxiv.org/pdf/1603.06560.pdf)

    """

    def predict_one(self, x):
        return self.best_model.predict_one(x)


class SuccessiveHalvingClassifier(SuccessiveHalving, base.Classifier):
    """Successive halving algorithm for classification.

    Successive halving is a method for performing model selection without having to train each
    model on all the dataset. At certain points in time (called "rungs"), the worst performing will
    be discarded and the best ones will keep competing between each other. The rung values are
    designed so that at most `budget` model updates will be performed in total.

    If you have `k` combinations of hyperparameters and that your dataset contains `n`
    observations, then the maximal budget you can allocate is:

    .. math:: \\frac{2kn}{eta}

    It is recommended that you check this beforehand. This bound can't be checked by the function
    because the size of the dataset is not known. In fact it is potentially infinite, in which case
    the algorithm will terminate once all the budget has been spent.

    If you have a budget of `B`, and that your dataset contains `n` observations, then the
    number of hyperparameter combinations that will spend all the budget and go through all the
    data is:

    .. math:: \\ceil(\\floor(\\frac{B}{(2n)}) \times eta)

    Parameters:
        models: The models to compare.
        X_y: A stream of (features, target) tuples.
        metric: Metric used for comparing models with.
        budget: Total number of model updates you wish to allocate.
        eta: Rate of elimination. At every rung, `math.ceil(k / eta)` models are kept, where
            `k` is the number of models that have reached the rung. A higher `eta` value will
            focus on less models but will allocate more iterations to the best models.
        verbose: Whether to display progress or not.
        print_kwargs: Extra keyword arguments are passed to the `print` function. For instance,
            this allows providing a `file` argument, which indicates where to output progress.

    Examples:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression()
        ... )

        >>> param_grid = model_selection.expand_param_grid({
        ...     'LogisticRegression': {
        ...         'optimizer': [
        ...             (optim.SGD, {'lr': [.1, .01, .005]}),
        ...             (optim.Adam, {'beta_1': [.01, .001], 'lr': [.1, .01, .001]}),
        ...             (optim.Adam, {'beta_1': [.1], 'lr': [.001]}),
        ...         ]
        ...     }
        ... })

        >>> models = [model._set_params(params) for params in param_grid]

        >>> sh = model_selection.SuccessiveHalvingClassifier(
        ...     models=models,
        ...     metric=metrics.Accuracy(),
        ...     budget=2000,
        ...     eta=2,
        ...     verbose=True
        ... )

        >>> model_selection.progressive_val_score(
        ...     X_y=datasets.Phishing(),
        ...     model=sh,
        ...     metric=metrics.ROCAUC()
        ... )
        [1]	5 removed	5 left	50 iterations	budget used: 500	budget left: 1500	best Accuracy: 80.00%
        [2]	2 removed	3 left	100 iterations	budget used: 1000	budget left: 1000	best Accuracy: 84.00%
        [3]	1 removed	2 left	166 iterations	budget used: 1498	budget left: 502	best Accuracy: 86.14%
        [4]	1 removed	1 left	250 iterations	budget used: 1998	budget left: 2	best Accuracy: 84.80%
        ROCAUC: 0.953198

        >>> sh.best_model
        Pipeline (
          StandardScaler (),
          LogisticRegression (
            optimizer=Adam (
              lr=Constant (
                learning_rate=0.01
              )
              beta_1=0.01
              beta_2=0.999
              eps=1e-08
            )
            loss=Log (
              weight_pos=1.
              weight_neg=1.
            )
            l2=0.
            intercept=-0.399002
            intercept_lr=Constant (
              learning_rate=0.01
            )
            clip_gradient=1e+12
            initializer=Zeros ()
          )
        )

    References:
        1. [Jamieson, K. and Talwalkar, A., 2016, May. Non-stochastic best arm identification and hyperparameter optimization. In Artificial Intelligence and Statistics (pp. 240-248).](http://proceedings.mlr.press/v51/jamieson16.pdf)
        2. [Li, L., Jamieson, K., Rostamizadeh, A., Gonina, E., Hardt, M., Recht, B. and Talwalkar, A., 2018. Massively parallel hyperparameter tuning. arXiv preprint arXiv:1810.05934.](https://arxiv.org/pdf/1810.05934.pdf)
        3. [Li, L., Jamieson, K., DeSalvo, G., Rostamizadeh, A. and Talwalkar, A., 2017. Hyperband: A novel bandit-based approach to hyperparameter optimization. The Journal of Machine Learning Research, 18(1), pp.6765-6816.](https://arxiv.org/pdf/1603.06560.pdf)

    """

    def predict_proba_one(self, x):
        return self.best_model.predict_proba_one(x)



import collections
import functools
import itertools
import math
import operator
import re
import typing
import unicodedata

from creme import base


__all__ = ['BagOfWords', 'TFIDF']


N_GRAM = typing.Union[
    str,  # unigram
    typing.Tuple[str, ...]  # n-gram
]


def strip_accents_unicode(s: str) -> str:
    """Transform accentuated unicode symbols into their ASCII counterpart."""
    try:
        # If `s` is ASCII-compatible, then it does not contain any accented
        # characters and we can avoid an expensive list comprehension
        s.encode('ASCII', errors='strict')
        return s
    except UnicodeEncodeError:
        normalized = unicodedata.normalize('NFKD', s)
        return ''.join([c for c in normalized if not unicodedata.combining(c)])



def find_ngrams(tokens: typing.List[str], n: int) -> typing.Iterator[N_GRAM]:
    """Generates n-grams from a list of tokens.

    From http://www.locallyoptimal.com/blog/2013/01/20/elegant-n-gram-generation-in-python/.

    Example:

        >>> tokens = ['a', 'b', 'c']

        >>> list(find_ngrams(tokens, 0))
        []

        >>> list(find_ngrams(tokens, 1))
        ['a', 'b', 'c']

        >>> list(find_ngrams(tokens, 2))
        [('a', 'b'), ('b', 'c')]

        >>> list(find_ngrams(tokens, 3))
        [('a', 'b', 'c')]

        >>> list(find_ngrams(tokens, 4))
        []

    """
    if n == 1:
        return iter(tokens)
    return zip(*[tokens[i:] for i in range(n)])


def find_all_ngrams(tokens: typing.List[str], ngram_range: range) -> typing.Iterator[N_GRAM]:
    """Generates all n-grams in a given range.

    Example:

        >>> tokens = ['a', 'b', 'c']

        >>> list(find_all_ngrams(tokens, range(2)))
        ['a', 'b', 'c']

        >>> list(find_all_ngrams(tokens, range(1, 4)))
        ['a', 'b', 'c', ('a', 'b'), ('b', 'c'), ('a', 'b', 'c')]

    """
    return itertools.chain(*(find_ngrams(tokens, n) for n in ngram_range))


class VectorizerMixin:
    """Contains common processing steps used by each vectorizer.

    Parameters:
        on: The name of the feature that contains the text to vectorize. If `None`, then each
            `fit_one` and `transform_one` should treat `x` as a `str` and not as a `dict`.
        strip_accents: Whether or not to strip accent characters.
        lowercase: Whether or not to convert all characters to lowercase.
        preprocessor: Override the preprocessing step while preserving the tokenizing
            and n-grams generation steps.
        tokenizer: A function used to convert preprocessed text into a `dict` of tokens.
            A default tokenizer is used if `None` is passed. Set to `False` to disable
            tokenization.
        ngram_range: The lower and upper boundary of the range n-grams to be
            extracted. All values of n such that `min_n <= n <= max_n` will be used. For example
            an `ngram_range` of `(1, 1)` means only unigrams, `(1, 2)` means unigrams and
            bigrams, and `(2, 2)` means only bigrams.

    Attributes:
        processing_steps (list): A list of preprocessing steps that are applied to each text.

    """

    def __init__(self, on: str = None, strip_accents=True, lowercase=True,
                 preprocessor: typing.Callable = None,
                 tokenizer: typing.Callable = None, ngram_range=(1, 1)):
        self.on = on
        self.strip_accents = strip_accents
        self.lowercase = lowercase
        self.preprocessor = preprocessor
        self.tokenizer = re.compile(r'(?u)\b\w\w+\b').findall if tokenizer is None else tokenizer
        self.ngram_range = ngram_range

        self.processing_steps = []

        # Text extraction

        if on is not None:
            self.processing_steps.append(operator.itemgetter(on))

        # Preprocessing

        if preprocessor is not None:
            self.processing_steps.append(preprocessor)
        else:
            if self.strip_accents:
                self.processing_steps.append(strip_accents_unicode)
            if self.lowercase:
                self.processing_steps.append(str.lower)

        # Tokenization

        if self.tokenizer:
            self.processing_steps.append(self.tokenizer)

        if ngram_range[1] > 1:
            self.processing_steps.append(functools.partial(
                find_all_ngrams,
                ngram_range=range(ngram_range[0], ngram_range[1] + 1)
            ))

    def process_text(self, x):
        for step in self.processing_steps:
            x = step(x)
        return x

    def _more_tags(self):
        return {'handles_text': True}


class BagOfWords(base.Transformer, VectorizerMixin):
    """Counts tokens in sentences.

    This transformer can be used to counts tokens in a given piece of text. It takes care of
    normalizing the text before tokenizing it.

    Note that the parameters are identical to those of `feature_extraction.TFIDF`.

    Parameters:
        on: The name of the feature that contains the text to vectorize. If `None`, then each
            `fit_one` and `transform_one` will assume that each `x` that is provided is a `str`,
            andnot a `dict`.
        strip_accents: Whether or not to strip accent characters.
        lowercase: Whether or not to convert all characters to lowercase.
        preprocessor: Override the preprocessing step while preserving the tokenizing and n-grams
            generation steps.
        tokenizer: A function used to convert preprocessed text into a `dict` of tokens. By
            default, a regex formula that works well in most cases is used.
        ngram_range: The lower and upper boundary of the range n-grams to be
            extracted. All values of n such that `min_n <= n <= max_n` will be used. For example
            an `ngram_range` of `(1, 1)` means only unigrams, `(1, 2)` means unigrams and
            bigrams, and `(2, 2)` means only bigrams.

    Example:

        By default, `BagOfWords` will take as input a sentence, preprocess it, tokenize the
        preprocessed text, and then return a `collections.Counter` containing the number of
        occurrences of each token.

        >>> import creme

        >>> corpus = [
        ...     'This is the first document.',
        ...     'This document is the second document.',
        ...     'And this is the third one.',
        ...     'Is this the first document?',
        ... ]

        >>> bow = creme.feature_extraction.BagOfWords()

        >>> for sentence in corpus:
        ...     print(bow.transform_one(sentence))
        Counter({'this': 1, 'is': 1, 'the': 1, 'first': 1, 'document': 1})
        Counter({'document': 2, 'this': 1, 'is': 1, 'the': 1, 'second': 1})
        Counter({'and': 1, 'this': 1, 'is': 1, 'the': 1, 'third': 1, 'one': 1})
        Counter({'is': 1, 'this': 1, 'the': 1, 'first': 1, 'document': 1})

        Note that `fit_one` does not have to be called because `BagOfWords` is stateless. You can
        call it but it won't do anything.

        In the above example, a string is passed to `transform_one`. You can also indicate which
        field to access if the string is stored in a dictionary:

        >>> bow = creme.feature_extraction.BagOfWords(on='sentence')

        >>> for sentence in corpus:
        ...     x = {'sentence': sentence}
        ...     print(bow.transform_one(x))
        Counter({'this': 1, 'is': 1, 'the': 1, 'first': 1, 'document': 1})
        Counter({'document': 2, 'this': 1, 'is': 1, 'the': 1, 'second': 1})
        Counter({'and': 1, 'this': 1, 'is': 1, 'the': 1, 'third': 1, 'one': 1})
        Counter({'is': 1, 'this': 1, 'the': 1, 'first': 1, 'document': 1})

        The `ngram_range` parameter can be used to extract n-grams (including unigrams):

        >>> ngrammer = creme.feature_extraction.BagOfWords(ngram_range=(1, 2))

        >>> ngrams = ngrammer.transform_one('I love the smell of napalm in the morning')
        >>> for ngram, count in ngrams.items():
        ...     print(ngram, count)
        love 1
        the 2
        smell 1
        of 1
        napalm 1
        in 1
        morning 1
        ('love', 'the') 1
        ('the', 'smell') 1
        ('smell', 'of') 1
        ('of', 'napalm') 1
        ('napalm', 'in') 1
        ('in', 'the') 1
        ('the', 'morning') 1

    """

    def transform_one(self, x):
        return collections.Counter(self.process_text(x))


class TFIDF(BagOfWords):
    """Computes TF-IDF values from sentences.

    The TF-IDF formula is the same one as scikit-learn. The only difference is the fact that the
    document frequencies are determined online, whereas in a batch setting they can be determined
    by performing an initial pass through the data.

    Note that the parameters are identical to those of `feature_extraction.BagOfWords`.

    Parameters:
        normalize: Whether or not the TF-IDF values by their L2 norm.
        on: The name of the feature that contains the text to vectorize. If `None`, then
            the input is treated as a document instead of a set of features.
        strip_accents: Whether or not to strip accent characters.
        lowercase: Whether or not to convert all characters to lowercase.
        preprocessor: Override the preprocessing step while preserving the tokenizing and n-grams
            generation steps.
        tokenizer: A function used to convert preprocessed text into a `dict` of tokens. By
            default, a regex formula that works well in most cases is used.
        ngram_range: The lower and upper boundary of the range n-grams to be
            extracted. All values of n such that `min_n <= n <= max_n` will be used. For example
            an `ngram_range` of `(1, 1)` means only unigrams, `(1, 2)` means unigrams and
            bigrams, and `(2, 2)` means only bigrams. Only works if `tokenizer` is not set to
            `False`.

    Attributes:
        dfs (collections.defaultdict): Document counts.
        n (int): Number of scanned documents.

    Example:

        >>> from creme import feature_extraction

        >>> tfidf = feature_extraction.TFIDF()

        >>> corpus = [
        ...     'This is the first document.',
        ...     'This document is the second document.',
        ...     'And this is the third one.',
        ...     'Is this the first document?',
        ... ]

        >>> for sentence in corpus:
        ...     tfidf = tfidf.fit_one(sentence)
        ...     print(tfidf.transform_one(sentence))
        {'this': 0.447, 'is': 0.447, 'the': 0.447, 'first': 0.447, 'document': 0.447}
        {'this': 0.333, 'document': 0.667, 'is': 0.333, 'the': 0.333, 'second': 0.469}
        {'and': 0.497, 'this': 0.293, 'is': 0.293, 'the': 0.293, 'third': 0.497, 'one': 0.497}
        {'is': 0.384, 'this': 0.384, 'the': 0.384, 'first': 0.580, 'document': 0.469}

        In the above example, a string is passed to `transform_one`. You can also indicate which
        field to access if the string is stored in a dictionary:

        >>> tfidf = feature_extraction.TFIDF(on='sentence')

        >>> for sentence in corpus:
        ...     x = {'sentence': sentence}
        ...     tfidf = tfidf.fit_one(x)
        ...     print(tfidf.transform_one(x))
        {'this': 0.447, 'is': 0.447, 'the': 0.447, 'first': 0.447, 'document': 0.447}
        {'this': 0.333, 'document': 0.667, 'is': 0.333, 'the': 0.333, 'second': 0.469}
        {'and': 0.497, 'this': 0.293, 'is': 0.293, 'the': 0.293, 'third': 0.497, 'one': 0.497}
        {'is': 0.384, 'this': 0.384, 'the': 0.384, 'first': 0.580, 'document': 0.469}

    """

    def __init__(self, normalize=True, on: str = None, strip_accents=True, lowercase=True,
                 preprocessor: typing.Callable = None, tokenizer: typing.Callable = None,
                 ngram_range=(1, 1)):
        super().__init__(
            on=on,
            strip_accents=strip_accents,
            lowercase=lowercase,
            preprocessor=preprocessor,
            tokenizer=tokenizer,
            ngram_range=ngram_range
        )
        self.normalize = normalize
        self.dfs = collections.Counter()
        self.n = 0

    def fit_one(self, x):

        # Update the document counts
        terms = self.process_text(x)
        self.dfs.update(set(terms))

        # Increment the global document counter
        self.n += 1

        return self

    def transform_one(self, x):

        term_counts = super().transform_one(x)
        n_terms = sum(term_counts.values())

        tfidfs = {}

        for term, count in term_counts.items():
            tf = count / n_terms
            idf = math.log(((1 + self.n) / (1 + self.dfs[term]))) + 1
            tfidfs[term] = tf * idf

        if self.normalize:
            norm = math.sqrt(sum(tfidf ** 2 for tfidf in tfidfs.values()))
            return {term: tfidf / norm for term, tfidf in tfidfs.items()}
        return tfidfs



import collections

from . import base


__all__ = ['AdaDelta']


class AdaDelta(base.Optimizer):
    """AdaDelta optimizer.

    Parameters:
        rho
        eps

    Attributes:
        g2 (collections.defaultdict)
        s2 (collections.defaultdict)

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.AdaDelta()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.805611

    References:
        1. [Zeiler, M.D., 2012. Adadelta: an adaptive learning rate method. arXiv preprint arXiv:1212.5701.](https://arxiv.org/pdf/1212.5701.pdf)

    """

    def __init__(self, rho=0.95, eps=1e-8):
        super().__init__(lr=None)
        self.rho = rho
        self.eps = eps
        self.g2 = collections.defaultdict(float)
        self.s2 = collections.defaultdict(float)

    def _rms(self, x):
        return (x + self.eps) ** 0.5

    def _update_after_pred(self, w, g):

        for i, gi in g.items():

            # Accumulate the gradient
            self.g2[i] = self.rho * self.g2[i] + (1 - self.rho) * gi ** 2

            # Compute the update
            step = - self._rms(self.s2[i]) / self._rms(self.g2[i]) * gi

            # Accumulate the update
            self.s2[i] = self.rho * self.s2[i] + (1 - self.rho) * step ** 2

            # Apply the update
            w[i] += step

        return w



import collections

from . import base
from . import mean
from . import summing


class Cov(base.Bivariate):
    """Covariance.

    Parameters:
        ddof: Delta Degrees of Freedom.

    Example:

        >>> from creme import stats

        >>> x = [-2.1,  -1,  4.3]
        >>> y = [   3, 1.1, 0.12]

        >>> cov = stats.Cov()

        >>> for xi, yi in zip(x, y):
        ...     print(cov.update(xi, yi).get())
        0.0
        -1.044999
        -4.286

    References:
        1. [Wikipedia article on algorithms for calculating variance](https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Covariance)

    """

    def __init__(self, ddof=1):
        self.ddof = ddof
        self.mean_x = mean.Mean()
        self.mean_y = mean.Mean()
        self.cov = 0

    def update(self, x, y):
        dx = x - self.mean_x.get()
        self.mean_x.update(x)
        self.mean_y.update(y)
        dy = y - self.mean_y.get()
        self.cov += (dx * dy - self.cov) / max(1, self.mean_x.n - self.ddof)
        return self

    def get(self):
        return self.cov


class RollingCov(base.Bivariate):
    """Rolling covariance.

    Parameters:
        window_size: Size of the window over which to compute the covariance.
        ddof: Delta Degrees of Freedom.

    Here is the derivation, where $C$ denotes the covariance and $d$ is the amount of degrees of
    freedom:

    $$C = \frac{1}{n - d} \sum_{i=1}^n (x_i - \bar{x}) (y_i - \bar{y})$$

    $$C = \frac{1}{n - d} \sum_{i=1}^n x_i y_i - x_i \bar{y} - \bar{x} y_i + \bar{x} \bar{y}$$

    $$C = \frac{1}{n - d} (\sum_{i=1}^n x_i y_i - \bar{y} \sum_{i=1}^n x_i - \bar{x} \sum_{i=1}^n y_i + \sum_{i=1}^n \bar{x}\bar{y})$$

    $$C = \frac{1}{n - d} (\sum_{i=1}^n x_i y_i - \bar{y} n \bar{x} - \bar{x} n \bar{y} + n \bar{x}\bar{y})$$

    $$C = \frac{1}{n - d} (\sum_{i=1}^n x_i y_i - n \bar{x} \bar{y})$$

    $$C = \frac{1}{n - d} (\sum_{i=1}^n x_i y_i - \frac{\sum_{i=1}^n x_i \sum_{i=1}^n y_i}{n})$$

    The derivation is straighforward and somewhat trivial, but is a nice example of reformulating
    an equation so that it can be updated online. Note that we cannot apply this derivation to the
    non-rolling version of covariance because that would result in sums that grow infinitely, which
    can potentially cause numeric overflow.

    Example:

        >>> from creme import stats

        >>> x = [-2.1,  -1, 4.3, 1, -2.1,  -1, 4.3]
        >>> y = [   3, 1.1, .12, 1,    3, 1.1, .12]

        >>> rcov = stats.RollingCov(3)

        >>> for xi, yi in zip(x, y):
        ...     print(rcov.update(xi, yi).get())
        0.0
        -1.045
        -4.286
        -1.382
        -4.589
        -1.415
        -4.286

    """

    def __init__(self, window_size, ddof=1):
        self.ddof = ddof
        self.sx = summing.RollingSum(window_size)
        self.sy = summing.RollingSum(window_size)
        self.sxy = summing.RollingSum(window_size)

    @property
    def window_size(self):
        return self.sxy.window_size

    def update(self, x, y):
        self.sx.update(x)
        self.sy.update(y)
        self.sxy.update(x * y)
        return self

    def get(self):
        n = len(self.sx)  # current window size
        return (self.sxy.get() - self.sx.get() * self.sy.get() / n) / max(1, n - self.ddof)



import collections
import copy
import functools
import typing

import numpy as np

from creme import optim
from creme import utils

from . import base


__all__ = ['FunkMF']


class FunkMF(base.Recommender):
    """Funk Matrix Factorization for recommender systems.

    The model equation is defined as:

    $$\\hat{y}(x) = \\langle \\mathbf{v}_u, \\mathbf{v}_i \\rangle = \\sum_{f=1}^{k} \\mathbf{v}_{u, f} \\cdot \\mathbf{v}_{i, f}$$

    Where :math:`k` is the number of latent factors. The model expect dict inputs containing both a
    `user` and an `item` entries.

    Parameters:
        n_factors: Dimensionality of the factorization or number of latent factors.
        optimizer: The sequential optimizer used for updating the latent factors.
        loss: The loss function to optimize for.
        l2: Amount of L2 regularization used to push weights towards 0.
        initializer: Latent factors initialization scheme.
        clip_gradient: Clips the absolute value of each gradient value.
        seed: Randomization seed used for reproducibility.

    Attributes:
        u_latents (collections.defaultdict): The user latent vectors randomly initialized.
        i_latents (collections.defaultdict): The item latent vectors randomly initialized.
        u_optimizer (optim.Optimizer): The sequential optimizer used for updating the user latent
            weights.
        i_optimizer (optim.Optimizer): The sequential optimizer used for updating the item latent
            weights.

    Example:

        >>> from creme import optim
        >>> from creme import reco

        >>> X_y = (
        ...     ({'user': 'Alice', 'item': 'Superman'}, 8),
        ...     ({'user': 'Alice', 'item': 'Terminator'}, 9),
        ...     ({'user': 'Alice', 'item': 'Star Wars'}, 8),
        ...     ({'user': 'Alice', 'item': 'Notting Hill'}, 2),
        ...     ({'user': 'Alice', 'item': 'Harry Potter'}, 5),
        ...     ({'user': 'Bob', 'item': 'Superman'}, 8),
        ...     ({'user': 'Bob', 'item': 'Terminator'}, 9),
        ...     ({'user': 'Bob', 'item': 'Star Wars'}, 8),
        ...     ({'user': 'Bob', 'item': 'Notting Hill'}, 2)
        ... )

        >>> model = reco.FunkMF(
        ...     n_factors=10,
        ...     optimizer=optim.SGD(0.1),
        ...     initializer=optim.initializers.Normal(mu=0., sigma=0.1, seed=11),
        ... )

        >>> for x, y in X_y:
        ...     _ = model.fit_one(x, y)

        >>> model.predict_one({'user': 'Bob', 'item': 'Harry Potter'})
        1.866272

    .. note::
        This model expects a dict input with a `user` and an `item` entries without any type
        constraint on their values (i.e. can be strings or numbers). Other entries are ignored.

    References:
        1. [Netflix update: Try this at home](https://sifter.org/simon/journal/20061211.html)
        2. [Matrix factorization techniques for recommender systems](https://datajobs.com/data-science-repo/Recommender-Systems-[Netflix].pdf)

    """

    def __init__(self, n_factors=10, optimizer: optim.Optimizer = None,
                 loss: optim.losses.Loss = None, l2=0.,
                 initializer: optim.initializers.Initializer = None,
                 clip_gradient=1e12, seed: int = None):

        self.n_factors = n_factors
        self.u_optimizer = optim.SGD() if optimizer is None else copy.deepcopy(optimizer)
        self.i_optimizer = optim.SGD() if optimizer is None else copy.deepcopy(optimizer)
        self.loss = optim.losses.Squared() if loss is None else loss
        self.l2 = l2

        if initializer is None:
            initializer = optim.initializers.Normal(mu=0., sigma=.1, seed=seed)
        self.initializer = initializer

        self.clip_gradient = clip_gradient
        self.seed = seed

        random_latents = functools.partial(
            self.initializer,
            shape=self.n_factors
        )
        self.u_latents: typing.DefaultDict[int, optim.initializers.Initializer] = collections.defaultdict(random_latents)
        self.i_latents: typing.DefaultDict[int, optim.initializers.Initializer] = collections.defaultdict(random_latents)

    def _predict_one(self, user, item):
        return np.dot(self.u_latents[user], self.i_latents[item])

    def _fit_one(self, user, item, y):

        # Calculate the gradient of the loss with respect to the prediction
        g_loss = self.loss.gradient(y, self._predict_one(user, item))

        # Clamp the gradient to avoid numerical instability
        g_loss = utils.math.clamp(g_loss, minimum=-self.clip_gradient, maximum=self.clip_gradient)

        # Calculate latent gradients
        u_latent_grad = {user: g_loss * self.i_latents[item] + self.l2 * self.u_latents[user]}
        i_latent_grad = {item: g_loss * self.u_latents[user] + self.l2 * self.i_latents[item]}

        # Update latent weights
        self.u_latents = self.u_optimizer.update_after_pred(self.u_latents, u_latent_grad)
        self.i_latents = self.i_optimizer.update_after_pred(self.i_latents, i_latent_grad)

        return self



import abc
import collections
import functools
import io
import itertools
import math
import numbers
import typing

try:
    import graphviz
    GRAPHVIZ_INSTALLED = True
except ImportError:
    GRAPHVIZ_INSTALLED = False

from creme import base
from creme import proba

from . import criteria
from . import node
from . import splitting

CRITERIA_CLF = {'gini': criteria.gini_impurity, 'entropy': criteria.entropy}


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..., (s3, None)

    We use this to iterate over successive pairs of nodes in a path for curtailment purposes.

    """
    a, b = itertools.tee(iterable)
    next(b, None)
    return itertools.zip_longest(a, b)


class BaseDecisionTree(abc.ABC):

    def __init__(self, criterion='gini', patience=250, max_depth=5, min_split_gain=0.,
                 min_child_samples=20, confidence=1e-10, tie_threshold=5e-2, n_split_points=30,
                 max_bins=60, curtail_under=10):
        self.criterion = criterion
        self.patience = patience
        self.max_depth = max_depth
        self.min_split_gain = min_split_gain
        self.min_child_samples = min_child_samples
        self.confidence = confidence
        self.tie_threshold = tie_threshold
        self.n_split_points = n_split_points
        self.max_bins = max_bins
        self.curtail_under = curtail_under

        self.criterion_func = CRITERIA_CLF[criterion]
        self.root = node.Leaf(depth=0, tree=self, target_dist=proba.Multinomial())

    def fit_one(self, x, y):
        self.root = self.root.update(x, y)
        return self

    @abc.abstractmethod
    def _make_leaf_dist(self):
        """Returns a target distribution for a newly instantiated leaf."""

    @abc.abstractmethod
    def _get_split_enum(self, value):
        """Returns the appropriate split enumerator for a given value based on it's type."""

    def draw(self, max_depth: int = None):
        """Draws the tree using the `graphviz` library.

        Parameters:
            max_depth: Only the root will be drawn when set to `0`. Every node will be drawn when
                set to `None`.

        Example:

            >>> from creme import datasets
            >>> from creme import tree

            >>> model = tree.DecisionTreeClassifier(
            ...    patience=10,
            ...    confidence=1e-5,
            ...    criterion='gini',
            ...    max_depth=10,
            ...    tie_threshold=0.05,
            ...    min_child_samples=0,
            ... )

            >>> for x, y in datasets.Phishing():
            ...    model = model.fit_one(x, y)

            >>> dot = model.draw()

        .. image:: /img/dtree_draw.svg
            :align: center

        """

        if max_depth is None:
            max_depth = math.inf

        dot = graphviz.Digraph(
            graph_attr={'splines': 'ortho'},
            node_attr={
                'shape': 'box', 'penwidth': '1.2', 'fontname': 'trebuchet',
                'fontsize': '11', 'margin': '0.1,0.0'
            },
            edge_attr={'penwidth': '0.6', 'center': 'true'}
        )

        # Do a first pass to guess the number of classes
        n_classes = len(set(itertools.chain(*[
            list(node.target_dist)
            for node, _ in self.root.iter_dfs()])
        ))

        # Pick a color palette which maps classes to colors
        new_color = functools.partial(next, iter(_color_brew(n_classes)))
        palette = collections.defaultdict(new_color)

        for parent_no, child_no, _, child, child_depth in self.root.iter_edges():

            if child_depth > max_depth:
                continue

            if isinstance(child, node.Branch):
                text = f'{child.split} \n {child.target_dist} \n samples: {child.n_samples}'
            elif isinstance(child, node.Leaf):
                text = f'{child.target_dist} \n samples: {child.n_samples}'

            # Pick a color, the hue depends on the class and the transparency on the distribution
            mode = child.target_dist.mode
            if mode is not None:
                p_mode = child.target_dist.pmf(mode)
                alpha = (p_mode - 1 / n_classes) / (1 - 1 / n_classes)
                fillcolor = str(transparency_hex(color=palette[mode], alpha=alpha))
            else:
                fillcolor = '#FFFFFF'

            dot.node(f'{child_no}', text, fillcolor=fillcolor, style='filled')

            if parent_no is not None:
                dot.edge(f'{parent_no}', f'{child_no}')

        return dot

    def debug_one(self, x: dict) -> str:
        """Prints an explanation of how `x` is predicted.

        Parameters:
            x: A dictionary of features.

        """

        # We'll redirect all the print statement to a buffer, we'll return the content of the
        # buffer at the end
        buffer = io.StringIO()
        _print = functools.partial(print, file=buffer)

        node = self.root

        for node in self.root.path(x):
            if isinstance(node, node.Leaf):
                _print(node.target_dist)
                break
            if node.split(x):
                _print('not', node.split)
            else:
                _print(node.split)

        return buffer.getvalue()


class DecisionTreeClassifier(BaseDecisionTree, base.MultiClassifier):
    """Decision tree classifier.

    Numeric features will be treated as continuous features. You can cast a feature to a string in
    order to treat it as a categorical feature.

    Parameters:
        criterion: The function to measure the quality of a split. Set to `'gini'` in order
            to use Gini impurity and `'entropy'` for information gain.
        patience: Time to wait between split attempts.
        max_depth: Maximum tree depth.
        min_split_gain: Minimum impurity gain required to make a split eligible.
        min_child_samples: Minimum number of data needed in a leaf.
        confidence: Threshold used to compare with the Hoeffding bound.
        tie_threshold: Threshold to handle ties between equally performing attributes.
        n_split_points: Number of split points considered for splitting numerical variables.
        max_bins: Number of histogram bins used for approximating the distribution of
            numerical variables.
        curtail_under: Determines the minimum amount of samples for a node to be eligible to
            make predictions. For instance, if a leaf doesn't contain at least `curtail_under`
            samples, then it's parent will be used instead. If said parent also doesn't contain at
            leaf `curtail_under` samples, then it's parent is used, etc. This helps to counter
            the fact that new leaves start with no samples at all, therefore their predictions
            might be unreliable. No curtailment will be applied if you set this to `0`. However,
            note that using even a small amount of curtailment almost always results in better
            performance.

    Attributes:
        root

    Example:

        >>> from creme import datasets
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import tree

        >>> X_y = datasets.Phishing()

        >>> model = tree.DecisionTreeClassifier(
        ...     patience=100,
        ...     confidence=1e-5,
        ...     criterion='gini'
        ... )

        >>> metric = metrics.LogLoss()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        LogLoss: 0.51755

    References:
        1. [Domingos, P. and Hulten, G., 2000, August. Mining high-speed data streams. In Proceedings of the sixth ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 71-80).](https://homes.cs.washington.edu/~pedrod/papers/kdd00.pdf)
        2. [Article by The Morning Paper](https://blog.acolyer.org/2015/08/26/mining-high-speed-data-streams/)

    """

    def _make_leaf_dist(self):
        return proba.Multinomial()

    def _get_split_enum(self, value):
        """Returns an appropriate split enumerator given a feature's type."""
        if isinstance(value, numbers.Number):
            return splitting.HistSplitEnum(n_bins=self.max_bins, n_splits=self.n_split_points)

        elif isinstance(value, (str, bool)):
            return splitting.CategoricalSplitEnum()

        raise ValueError(f'The type of {value} ({type(value)}) is not supported')

    def predict_proba_one(self, x):

        # Find the deepest node which contains at least curtail_under samples
        for node, child in pairwise(self.root.path(x)):
            if child is None or child.n_samples < self.curtail_under:
                break

        return {c: node.target_dist.pmf(c) for c in node.target_dist}


def _color_brew(n: int) -> typing.List[typing.Tuple[int, int, int]]:
    """Generate n colors with equally spaced hues.

    Parameters:
        n: The number of colors required.

    Returns:
        List of n tuples of form (R, G, B) being the components of each color.

    References:
        https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/tree/_export.py

    """
    colors = []

    # Initialize saturation & value; calculate chroma & value shift
    s, v = .75, .9
    c = s * v
    m = v - c

    for h in [i for i in range(25, 385, int(360 / n))]:

        # Calculate some intermediate values
        h_bar = h / 60.
        x = c * (1 - abs((h_bar % 2) - 1))

        # Initialize RGB with same hue & chroma as our color
        rgb = [(c, x, 0),
               (x, c, 0),
               (0, c, x),
               (0, x, c),
               (x, 0, c),
               (c, 0, x),
               (c, x, 0)]
        r, g, b = rgb[int(h_bar)]

        # Shift the initial RGB values to match value and store
        colors.append((
            (int(255 * (r + m))),
            (int(255 * (g + m))),
            (int(255 * (b + m)))
        ))

    return colors


def transparency_hex(color: typing.Tuple[int, int, int], alpha: float) -> str:
    """Apply alpha coefficient on hexadecimal color."""
    return '#%02x%02x%02x' % tuple([int(round(alpha * c + (1 - alpha) * 255, 0)) for c in color])



import collections
import itertools
import math
import typing

from creme import linear_model
from creme import preprocessing
import creme.base

from . import base


__all__ = ['SNARIMAX']


def make_coeffs(d, m):
    """Precomputes the coefficients of the backshift operator.

    Example:

        >>> make_coeffs(1, 1)
        {0: -1}

        >>> make_coeffs(2, 1)
        {0: -2, 1: 1}

        >>> make_coeffs(3, 1)
        {0: -3, 1: 3, 2: -1}

        >>> make_coeffs(2, 7)
        {6: -2, 13: 1}

    """

    def n_choose_k(n, k):
        f = math.factorial
        return f(n) // f(k) // f(n - k)

    return dict(
        (
            k * m - 1,
            int(math.copysign(1, (k + 1) % 2 - 1)) * n_choose_k(n=d, k=k)
        )
        for k in range(1, d + 1)
    )


class Differencer:
    """A time series differencer.

    Example:

        >>> differencer = Differencer(2); differencer.coeffs
        {0: -2, 1: 1}

        >>> differencer.diff(7, [3, 1])
        2

        >>> differencer.undiff(2, [3, 1])
        7

    References:
        1. [Stationarity and differencing](https://otexts.com/fpp2/stationarity.html)

    """

    def __init__(self, d, m=1):

        if d < 0:
            raise ValueError('d must be greater than or equal to 0')

        if m < 1:
            raise ValueError('m must be greater than or equal to 1')

        self.coeffs = make_coeffs(d=d, m=m)

    def __add__(self, other):
        """Composes two differencers together.

        Example:

            >>> differencer = Differencer(d=3, m=2) + Differencer(d=3, m=1)
            >>> for t, c in sorted(differencer.coeffs.items()):
            ...     print(t, c)
            0 -3
            2 8
            3 -6
            4 -6
            5 8
            7 -3
            8 1

        References:
            1. [Backshift notation](https://otexts.com/fpp2/backshift.html)

        """
        coeffs = collections.Counter()
        coeffs.update(self.coeffs)
        coeffs.update(other.coeffs)

        for (t1, c1), (t2, c2) in itertools.product(self.coeffs.items(), other.coeffs.items()):
            coeffs[t1 + t2 + 1] += c1 * c2

        # Remove 0 coefficients
        for t in list(coeffs.keys()):
            if coeffs[t] == 0:
                del coeffs[t]

        differencer = Differencer(0, 1)
        differencer.coeffs = dict(coeffs)
        return differencer

    def diff(self, y, y_previous):
        """Differentiates a value.

            y (float): The value to differentiate.
            y_previous (list of float): The window of previous values. The first element is assumed
                to be the most recent value.

        """
        return y + sum(
            c * y_previous[t]
            for t, c in self.coeffs.items()
            if t < len(y_previous)
        )

    def undiff(self, y: float, y_previous: typing.List[float]):
        """Undifferentiates a value.

            y: The value to differentiate.
            y_previous: The window of previous values. The first element is assumed to be the most
                recent value.

        """
        return y - sum(
            c * y_previous[t]
            for t, c in self.coeffs.items()
            if t < len(y_previous)
        )


class SNARIMAX(base.Forecaster):
    """SNARIMAX model.

    SNARIMAX stands for (S)easonal (N)on-linear (A)uto(R)egressive (I)ntegrated (M)oving-(A)verage
    with e(X)ogenous inputs model.

    This model generalizes many established time series models in a single interface that can be
    trained online. It assumes that the provided training data is ordered in time and is uniformly
    spaced. It is made up of the following components:

    - S (Seasonal)
    - N (Non-linear): Any online regression model can be used, not necessarily a linear regression
        as is done in textbooks.
    - AR (Autoregressive): Lags of the target variable are used as features.
    - I (Integrated): The model can be fitted on a differenced version of a time series. In this
        context, integration is the reverse of differencing.
    - MA (Moving average): Lags of the errors are used as features.
    - X (Exogenous): Users can provide additional features. Care has to be taken to include
        features that will be available both at training and prediction time.

    Each of these components can be switched on and off by specifying the appropriate parameters.
    Classical time series models such as AR, MA, ARMA, and ARIMA can thus be seen as special
    parametrizations of the SNARIMAX model.

    Parameters:
        p: Order of the autoregressive part. This is the number of past target values that will be
            included as features.
        d: Differencing order.
        q: Order of the moving average part. This is the number of past error terms that will be
            included as features.
        m: Season length used for extracting seasonal features. If you believe your data has a
            seasonal pattern, then set this accordingly. For instance, if the data seems to exhibit
            a yearly seasonality, and that your data is spaced by month, then you should set this
            to 12. Note that for this parameter to have any impact you should also set at least one
            of the `p`, `d`, and `q` parameters.
        sp: Seasonal order of the autoregressive part. This is the number of past target values
            that will be included as features.
        sd: Seasonal differencing order.
        sq: Seasonal order of the moving average part. This is the number of past error terms that
            will be included as features.
        regressor: The online regression model to use. By default, a `preprocessing.StandardScaler`
            piped with a `linear_model.LinearRegression` will be used.

    Attributes:
        differencer (Differencer)
        y_trues (collections.deque): The `p` past target values.
        errors (collections.deque): The `q` past error values.

    Example:

        >>> import calendar
        >>> import datetime as dt
        >>> from creme import compose
        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import optim
        >>> from creme import preprocessing
        >>> from creme import time_series

        >>> def get_month_distances(x):
        ...     return {
        ...         calendar.month_name[month]: math.exp(-(x['month'].month - month) ** 2)
        ...         for month in range(1, 13)
        ...     }

        >>> def get_ordinal_date(x):
        ...     return {'ordinal_date': x['month'].toordinal()}

        >>> extract_features = compose.TransformerUnion(
        ...     get_ordinal_date,
        ...     get_month_distances
        ... )

        >>> model = (
        ...     extract_features |
        ...     time_series.SNARIMAX(
        ...         p=0,
        ...         d=0,
        ...         q=0,
        ...         m=12,
        ...         sp=3,
        ...         sq=6,
        ...         regressor=(
        ...             preprocessing.StandardScaler() |
        ...             linear_model.LinearRegression(
        ...                 intercept=110,
        ...                 optimizer=optim.SGD(0.01),
        ...                 intercept_lr=0.3
        ...             )
        ...         )
        ...     )
        ... )

        >>> metric = metrics.Rolling(metrics.MAE(), 12)

        >>> for x, y in datasets.AirlinePassengers():
        ...     y_pred = model.forecast(horizon=1, xs=[x])
        ...     model = model.fit_one(x, y)
        ...     metric = metric.update(y, y_pred[0])

        >>> metric
        Rolling of size 12 MAE: 11.636563

        >>> horizon = 12
        >>> future = [
        ...     {'month': dt.date(year=1961, month=m, day=1)}
        ...     for m in range(1, horizon + 1)
        ... ]
        >>> forecast = model.forecast(horizon=horizon, xs=future)
        >>> for x, y_pred in zip(future, forecast):
        ...     print(x['month'], f'{y_pred:.3f}')
        1961-01-01 442.554
        1961-02-01 427.305
        1961-03-01 471.861
        1961-04-01 483.978
        1961-05-01 489.995
        1961-06-01 544.270
        1961-07-01 632.882
        1961-08-01 633.229
        1961-09-01 531.349
        1961-10-01 457.258
        1961-11-01 405.978
        1961-12-01 439.674

    .. tip::
        This model is tailored for time series that are homoskedastic. In other words, it might not
        work well if the variance of the time series varies widely along time.

    References:
        1. [Wikipedia page on ARMA](https://www.wikiwand.com/en/Autoregressive%E2%80%93moving-average_model)
        2. [Wikipedia page on NARX](https://www.wikiwand.com/en/Nonlinear_autoregressive_exogenous_model)
        3. [ARIMA models](https://otexts.com/fpp2/arima.html)
        4. [Anava, O., Hazan, E., Mannor, S. and Shamir, O., 2013, June. Online learning for time series prediction. In Conference on learning theory (pp. 172-184)](https://arxiv.org/pdf/1302.6927.pdf)

    """

    def __init__(self, p: int, d: int, q: int, m: int = 1, sp: int = 0, sd: int = 0, sq: int = 0,
                 regressor: creme.base.Regressor = None):

        self.p = p
        self.d = d
        self.q = q
        self.m = m
        self.sp = sp
        self.sd = sd
        self.sq = sq
        self.regressor = (
            regressor if regressor is not None else
            preprocessing.StandardScaler() | linear_model.LinearRegression()
        )
        self.differencer = Differencer(d=d, m=1) + Differencer(d=sd, m=1)
        self.y_trues = collections.deque(maxlen=max(p, m * sp))
        self.errors = collections.deque(maxlen=max(p, m * sq))

    def _add_lag_features(self, x, y_trues, errors):

        if x is None:
            x = {}

        # AR
        for t in range(self.p):
            try:
                x[f'y-{t+1}'] = y_trues[t]
            except IndexError:
                break

        # Seasonal AR
        for t in range(self.m - 1, (self.m - 1) * self.sp, self.m):
            try:
                x[f'sy-{t+1}'] = y_trues[t]
            except IndexError:
                break

        # MA
        for t in range(self.q):
            try:
                x[f'e-{t+1}'] = errors[t]
            except IndexError:
                break

        # Seasonal MA
        for t in range(self.m - 1, (self.m - 1) * self.sq, self.m):
            try:
                x[f'se-{t+1}'] = errors[t]
            except IndexError:
                break

        return x

    def _fit_predict_one(self, y: float, x: dict = None):
        """Updates the model and returns the prediction for the next time step.

        Parameters:
            x: Optional additional features to learn from. In the litterature these are called the
                exogenous variables.
            y: In the litterature this is called the endogenous variable.

        """

        # Check there are enough observations so that differencing can happen
        y = self.differencer.diff(y=y, y_previous=self.y_trues)
        x = self._add_lag_features(x=x, y_trues=self.y_trues, errors=self.errors)
        y_pred = self.regressor.predict_one(x)
        self.regressor.fit_one(x, y)

        self.y_trues.appendleft(y)
        self.errors.appendleft(y - y_pred)

        return y_pred

    def fit_one(self, y, x=None):
        self._fit_predict_one(y=y, x=x)
        return self

    def forecast(self, horizon, xs=None):

        if xs is None:
            xs = [{}] * horizon

        if len(xs) != horizon:
            raise ValueError('the length of xs should be equal to the specified horizon')

        y_trues = collections.deque(self.y_trues)
        errors = collections.deque(self.errors)
        forecasts = [None] * horizon

        for t, x in enumerate(xs):
            x = self._add_lag_features(x=x, y_trues=y_trues, errors=errors)
            y_pred = self.regressor.predict_one(x)
            forecasts[t] = self.differencer.undiff(y=y_pred, y_previous=y_trues)

            y_trues.appendleft(y_pred)
            errors.appendleft(0)

        return forecasts



"""
creme is a library for incremental learning. Incremental learning is a machine learning regime
where the observations are made available one by one. It is also known as online learning,
iterative learning, or sequential learning. This is in contrast to batch learning where all the
data is processed at once. Incremental learning is desirable when the data is too big to fit in
memory, or simply when it isn't available all at once. creme's API is heavily inspired from that of
scikit-learn, enough so that users who are familiar with scikit-learn should feel right at home.
"""
from .__version__ import __version__

from . import anomaly
from . import base
from . import cluster
from . import compat
from . import compose
from . import datasets
from . import dummy
from . import decomposition
from . import ensemble
from . import facto
from . import feature_extraction
from . import feature_selection
from . import impute
from . import linear_model
from . import meta
from . import metrics
from . import model_selection
from . import multiclass
from . import multioutput
from . import naive_bayes
from . import neighbors
from . import optim
from . import preprocessing
from . import proba
from . import reco
from . import sampling
from . import stats
from . import stream
from . import time_series
from . import tree
from . import utils

__all__ = [
    'anomaly',
    'base',
    'cluster',
    'compat',
    'compose',
    'datasets',
    'dummy',
    'decomposition',
    'ensemble',
    'facto',
    'feature_extraction',
    'feature_selection',
    'impute',
    'linear_model',
    'meta',
    'metrics',
    'model_selection',
    'multiclass',
    'multioutput',
    'naive_bayes',
    'neighbors',
    'optim',
    'preprocessing',
    'proba',
    'reco',
    'sampling',
    'stats',
    'stream',
    'time_series',
    'tree',
    'utils'
]

__pdoc__ = {
    'anomaly': False,
    'cluster': False,
    'compat': False,
    'compose': False,
    'datasets': False,
    'dummy': False,
    'decomposition': False,
    'ensemble': False,
    'facto': False,
    'feature_extraction': False,
    'feature_selection': False,
    'impute': False,
    'linear_model': False,
    'meta': False,
    'multiclass': False,
    'multioutput': False,
    'naive_bayes': False,
    'optim': False,
    'preprocessing': False,
    'proba': False,
    'reco': False,
    'sampling': False,
    'time_series': False,
    'tree': False,
    'utils': False,

    'conftest': False
}



from creme import stream

from . import base


class Elec2(base.RemoteDataset):
    """Electricity prices in New South Wales.

    This is a binary classification task, where the goal is to predict if the price of electricity
    will go up or down.

    This data was collected from the Australian New South Wales Electricity Market. In this market,
    prices are not fixed and are affected by demand and supply of the market. They are set every
    five minutes. Electricity transfers to/from the neighboring state of Victoria were done to
    alleviate fluctuations.

    References:
        1. [SPLICE-2 Comparative Evaluation: Electricity Pricing](http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.12.9405)
        2. [DataHub description](https://datahub.io/machine-learning/electricity#readme)

    """

    def __init__(self):
        super().__init__(
            url='https://maxhalford.github.io/files/datasets/electricity.zip',
            size=3091689,
            task=base.BINARY_CLF,
            n_samples=45_312,
            n_features=8,
            filename='electricity.csv'
        )

    def _iter(self):
        return stream.iter_csv(
            self.path,
            target='class',
            converters={
                'date': float,
                'day': int,
                'period': float,
                'nswprice': float,
                'nswdemand': float,
                'vicprice': float,
                'vicdemand': float,
                'transfer': float,
                'class': lambda x: x == 'UP'
            }
        )



import collections
import copy
import functools
import importlib
import inspect
import math
import pickle
import random
import statistics

from creme import stats
import numpy as np
import pytest
from scipy import stats as sp_stats


def load_stats():
    for _, obj in inspect.getmembers(importlib.import_module('creme.stats'), inspect.isclass):
        try:

            if issubclass(obj, stats.Link):
                yield obj(stats.Shift(1), stats.Mean())
                continue

            sig = inspect.signature(obj)
            yield obj(**{
                param.name: param.default if param.default != param.empty else 1
                for param in sig.parameters.values()
            })
        except ValueError:
            yield obj()


@pytest.mark.parametrize('stat', load_stats(), ids=lambda stat: stat.__class__.__name__)
def test_pickling(stat):
    assert isinstance(pickle.loads(pickle.dumps(stat)), stat.__class__)
    assert isinstance(copy.deepcopy(stat), stat.__class__)

    # Check the statistic has a working __str__ and name method
    assert isinstance(str(stat), str)

    if isinstance(stat, stats.Univariate):
        assert isinstance(stat.name, str)


@pytest.mark.parametrize(
    'stat, func',
    [
        (stats.Kurtosis(bias=True), sp_stats.kurtosis),
        (stats.Kurtosis(bias=False), functools.partial(sp_stats.kurtosis, bias=False)),
        (stats.Mean(), statistics.mean),
        (stats.Skew(bias=True), sp_stats.skew),
        (stats.Skew(bias=False), functools.partial(sp_stats.skew, bias=False)),
        (stats.Var(ddof=0), np.var),
        (stats.Var(), functools.partial(np.var, ddof=1))
    ]
)
def test_univariate(stat, func):

    # Shut up
    np.warnings.filterwarnings('ignore')

    X = [random.random() for _ in range(30)]

    for i, x in enumerate(X):
        stat.update(x)
        if i >= 1:
            assert math.isclose(stat.get(), func(X[:i + 1]), abs_tol=1e-10)


@pytest.mark.parametrize(
    'stat, func',
    [
        (stats.RollingMean(3), statistics.mean),
        (stats.RollingMean(10), statistics.mean),
        (stats.RollingVar(3, ddof=0), np.var),
        (stats.RollingVar(10, ddof=0), np.var)
    ]
)
def test_rolling_univariate(stat, func):

    # We know what we're doing
    np.warnings.filterwarnings('ignore')

    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    n = stat.window_size
    X = [random.random() for _ in range(30)]

    for i, x in enumerate(X):
        stat.update(x)
        if i >= 1:
            assert math.isclose(stat.get(), func(tail(X[:i + 1], n)), abs_tol=1e-10)


@pytest.mark.parametrize(
    'stat, func',
    [
        (stats.Cov(), lambda x, y: np.cov(x, y)[0, 1]),
        (stats.PearsonCorr(), lambda x, y: sp_stats.pearsonr(x, y)[0])
    ]
)
def test_bivariate(stat, func):

    # Shhh
    np.warnings.filterwarnings('ignore')

    X = [random.random() for _ in range(30)]
    Y = [random.random() * x for x in X]

    for i, (x, y) in enumerate(zip(X, Y)):
        stat.update(x, y)
        if i >= 1:
            assert math.isclose(stat.get(), func(X[:i + 1], Y[:i + 1]), abs_tol=1e-10)


@pytest.mark.parametrize(
    'stat, func',
    [
        (stats.RollingPearsonCorr(3), lambda x, y: sp_stats.pearsonr(x, y)[0]),
        (stats.RollingPearsonCorr(10), lambda x, y: sp_stats.pearsonr(x, y)[0]),
        (stats.RollingCov(3), lambda x, y: np.cov(x, y)[0, 1]),
        (stats.RollingCov(10), lambda x, y: np.cov(x, y)[0, 1])
    ]
)
def test_rolling_bivariate(stat, func):

    # Enough already
    np.warnings.filterwarnings('ignore')

    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    n = stat.window_size
    X = [random.random() for _ in range(30)]
    Y = [random.random() * x for x in X]

    for i, (x, y) in enumerate(zip(X, Y)):
        stat.update(x, y)
        if i >= 1:
            x_tail = tail(X[:i + 1], n)
            y_tail = tail(Y[:i + 1], n)
            assert math.isclose(stat.get(), func(x_tail, y_tail), abs_tol=1e-10)



import collections
import math
import numbers
import typing

import numpy as np
import pandas as pd

from creme import base
from creme import optim
from creme import utils


__all__ = [
    'LinearRegression',
    'LogisticRegression'
]


class GLM:
    """Generalized Linear Model.

    This serves as a base class for linear and logistic regression.

    """

    def __init__(self, optimizer, loss, l2, intercept, intercept_lr, clip_gradient, initializer):
        self.optimizer = optimizer
        self.loss = loss
        self.l2 = l2
        self.intercept = intercept
        self.intercept_lr = (
            optim.schedulers.Constant(intercept_lr)
            if isinstance(intercept_lr, numbers.Number) else
            intercept_lr
        )
        self.clip_gradient = clip_gradient
        self.weights = collections.defaultdict(initializer)
        self.initializer = initializer

        # The predict_many functions are going to return pandas.Series. We can name the series with
        # the name given to the y series seen during the last fit_many call.
        self._y_name = None

    def _fit(self, x, y, w, get_grad):

        # Some optimizers need to do something before a prediction is made
        self.weights = self.optimizer.update_before_pred(w=self.weights)

        # Calculate the gradient
        gradient, loss_gradient = get_grad(x, y, w)

        # Update the intercept
        self.intercept -= self.intercept_lr.get(self.optimizer.n_iterations) * loss_gradient

        # Update the weights
        self.weights = self.optimizer.update_after_pred(w=self.weights, g=gradient)

        return self

    # Single instance methods

    def _raw_dot_one(self, x: dict) -> float:
        return utils.math.dot(self.weights, x) + self.intercept

    def _eval_gradient_one(self, x: dict, y: float, w: float) -> (dict, float):

        loss_gradient = self.loss.gradient(y_true=y, y_pred=self._raw_dot_one(x))
        loss_gradient *= w
        loss_gradient = np.clip(loss_gradient, -self.clip_gradient, self.clip_gradient)

        return (
            {
                i: xi * loss_gradient + 2. * self.l2 * self.weights.get(i, 0)
                for i, xi in x.items()
            },
            loss_gradient
        )

    def fit_one(self, x, y, w=1.):
        return self._fit(x, y, w, get_grad=self._eval_gradient_one)

    # Mini-batch methods

    def _raw_dot_many(self, X: pd.DataFrame) -> np.ndarray:
        weights = np.array([self.weights[c] for c in X.columns])
        return X.values @ weights + self.intercept

    def _eval_gradient_many(self,
                            X: pd.DataFrame,
                            y: pd.Series,
                            w: typing.Union[float, pd.Series]) -> (dict, float):

        loss_gradient = self.loss.gradient(y_true=y.values, y_pred=self._raw_dot_many(X))
        loss_gradient *= w
        loss_gradient = np.clip(loss_gradient, -self.clip_gradient, self.clip_gradient)

        # At this point we have a feature matrix X of shape (n, p). The loss gradient is a vector
        # of length p. We want to multiply each of X's by the corresponding value in the loss
        # gradient. When this is all done, we collapse X by computing the average of each column,
        # thereby obtaining the mean gradient of the batch. From thereon, the code reduces to the
        # single instance case.
        gradient = np.einsum('ij,i->ij', X.values, loss_gradient).mean(axis=0)

        return dict(zip(X.columns, gradient)), loss_gradient.mean()

    def fit_many(self, X: pd.DataFrame, y: pd.Series, w: typing.Union[float, pd.Series] = 1):
        self._y_name = y.name
        return self._fit(X, y, w, get_grad=self._eval_gradient_many)


class LinearRegression(GLM, base.Regressor):
    """Linear regression.

    This estimator supports learning with mini-batches. On top of the single instance methods, it
    provides the following methods: `fit_many`, `predict_many`, `predict_proba_many`. Each method
    takes as input a `pandas.DataFrame` where each column represents a feature.

    Parameters:
        optimizer: The sequential optimizer used for updating the weights. Note
            that the intercept is handled separately. Defaults to `optim.SGD(.01)`.
        loss: The loss function to optimize for. Defaults to `optim.losses.SquaredLoss`.
        l2: Amount of L2 regularization used to push weights towards 0.
        intercept: Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        l2: Amount of L2 regularization used to push weights towards 0.
        clip_gradient: Clips the absolute value of each gradient value.
        initializer: Weights initialization scheme.

    Attributes:
        weights (collections.defaultdict): The current weights.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import preprocessing

        >>> X_y = datasets.TrumpApproval()

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LinearRegression(intercept_lr=.1)
        ... )
        >>> metric = metrics.MAE()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MAE: 0.555971

        >>> model['LinearRegression'].intercept
        35.617670

        You can call the `debug_one` method to break down a prediction. This works even if the
        linear regression is part of a pipeline.

        >>> x, y = next(iter(X_y))
        >>> report = model.debug_one(x)
        >>> print(report)
        0. Input
        --------
        gallup: 43.84321 (float)
        ipsos: 46.19925 (float)
        morning_consult: 48.31875 (float)
        ordinal_date: 736389 (int)
        rasmussen: 44.10469 (float)
        you_gov: 43.63691 (float)
        <BLANKLINE>
        1. StandardScaler
        -----------------
        gallup: 1.18810 (float)
        ipsos: 2.10348 (float)
        morning_consult: 2.73545 (float)
        ordinal_date: -1.73032 (float)
        rasmussen: 1.26872 (float)
        you_gov: 1.48391 (float)
        <BLANKLINE>
        2. LinearRegression
        -------------------
        Name              Value      Weight      Contribution
            Intercept    1.00000    35.61767       35.61767
                ipsos    2.10348     0.62689        1.31866
        morning_consult    2.73545     0.24180        0.66144
                gallup    1.18810     0.43568        0.51764
            rasmussen    1.26872     0.28118        0.35674
                you_gov    1.48391     0.03123        0.04634
        ordinal_date   -1.73032     3.45162       -5.97242
        <BLANKLINE>
        Prediction: 32.54607

    .. tip::
        It is generally a good idea to use a `preprocessing.StandardScaler` to help the optimizer
        converge by scaling the input features.

    """

    def __init__(self, optimizer: optim.Optimizer = None, loss: optim.losses.RegressionLoss = None,
                 l2=.0, intercept=0.,
                 intercept_lr: typing.Union[optim.schedulers.Scheduler, float] = .01,
                 clip_gradient=1e+12, initializer: optim.initializers.Initializer = None):
        super().__init__(
            optimizer=optim.SGD(.01) if optimizer is None else optimizer,
            loss=optim.losses.Squared() if loss is None else loss,
            intercept=intercept,
            intercept_lr=intercept_lr,
            l2=l2,
            clip_gradient=clip_gradient,
            initializer=initializer if initializer else optim.initializers.Zeros()
        )

    def predict_one(self, x):
        return self.loss.mean_func(self._raw_dot_one(x))

    def predict_many(self, X):
        return pd.Series(
            self.loss.mean_func(self._raw_dot_many(X)),
            index=X.index,
            name=self._y_name,
            copy=False
        )

    def debug_one(self, x: dict, decimals=5) -> str:
        """Debugs the output of the linear regression.

        Parameters:
            x: A dictionary of features.
            decimals: The number of decimals use for printing each numeric value.

        Returns:
            A table which explains the output.

        """

        def fmt_float(x):
            return '{: ,.{prec}f}'.format(x, prec=decimals)

        names = list(map(str, x.keys())) + ['Intercept']
        values = list(map(fmt_float, list(x.values()) + [1]))
        weights = list(map(fmt_float, [self.weights.get(i, 0) for i in x] + [self.intercept]))
        contributions = [xi * self.weights.get(i, 0) for i, xi in x.items()] + [self.intercept]
        order = reversed(np.argsort(contributions))
        contributions = list(map(fmt_float, contributions))

        table = utils.pretty.print_table(
            headers=['Name', 'Value', 'Weight', 'Contribution'],
            columns=[names, values, weights, contributions],
            order=order
        )

        return table


class LogisticRegression(GLM, base.BinaryClassifier):
    """Logistic regression.

    This estimator supports learning with mini-batches. On top of the single instance methods, it
    provides the following methods: `fit_many`, `predict_many`, `predict_proba_many`. Each method
    takes as input a `pandas.DataFrame` where each column represents a feature.

    Parameters:
        optimizer: The sequential optimizer used for updating the weights. Note that the intercept
            is handled separately. Defaults to `optim.SGD(.05)`.
        loss: The loss function to optimize for. Defaults to `optim.losses.Log`.
        l2: Amount of L2 regularization used to push weights towards 0.
        intercept: Initial intercept value.
        intercept_lr: Learning rate scheduler used for updating the intercept. If a `float` is
            passed, then an instance of `optim.schedulers.Constant` will be used. Setting this to 0
            implies that the intercept will be not be updated.
        l2: Amount of L2 regularization used to push weights towards 0.
        clip_gradient: Clips the absolute value of each gradient value.
        initializer: Weights initialization scheme.

    Attributes:
        weights (collections.defaultdict): The current weights.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer=optim.SGD(.1))
        ... )

        >>> metric = metrics.Accuracy()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        Accuracy: 88.96%

    .. tip::
        It is generally a good idea to use a `preprocessing.StandardScaler` to help the optimizer
        converge by scaling the input features.

    """

    def __init__(self, optimizer: optim.Optimizer = None, loss: optim.losses.BinaryLoss = None,
                 l2=.0, intercept=0.,
                 intercept_lr: typing.Union[float, optim.schedulers.Scheduler] = .01,
                 clip_gradient=1e12, initializer: optim.initializers.Initializer = None):

        super().__init__(
            optimizer=optim.SGD(.01) if optimizer is None else optimizer,
            loss=optim.losses.Log() if loss is None else loss,
            intercept=intercept,
            intercept_lr=intercept_lr,
            l2=l2,
            clip_gradient=clip_gradient,
            initializer=initializer if initializer else optim.initializers.Zeros()
        )

    def predict_proba_one(self, x):
        p = self.loss.mean_func(self._raw_dot_one(x))  # Convert logit to probability
        return {False: 1. - p, True: p}

    def predict_proba_many(self, X: pd.DataFrame) -> pd.DataFrame:
        p = self.loss.mean_func(self._raw_dot_many(X))  # Convert logits to probabilities
        return pd.DataFrame({False: 1. - p, True: p}, index=X.index, copy=False)

    def predict_many(self, X: pd.DataFrame) -> pd.Series:
        p = self.loss.mean_func(self._raw_dot_many(X))
        return pd.Series(p > .5, name=self._y_name, index=X.index, copy=False)



from creme import base

from . import utils


def iter_libsvm(filepath_or_buffer: str, target_type=float, compression='infer') -> base.typing.Stream:
    """Iterates over a dataset in LIBSVM format.

    The LIBSVM format is a popular way in the machine learning community to store sparse datasets.
    Only numerical feature values are supported. The feature names will be considered as strings.

    Parameters:
        filepath_or_buffer: Either a string indicating the location of a CSV file, or a buffer
            object that has a `read` method.
        target_type: The type of the target value.
        compression: For on-the-fly decompression of on-disk data. If this is set to 'infer' and
            `filepath_or_buffer` is a path, then the decompression method is inferred for the
            following extensions: '.gz', '.zip'.

    Example:

        >>> import io
        >>> from creme import stream

        >>> data = io.StringIO('''+1 x:-134.26 y:0.2563
        ... 1 x:-12 z:0.3
        ... -1 y:.25
        ... ''')

        >>> for x, y in stream.iter_libsvm(data, target_type=int):
        ...     print(y, x)
        1 {'x': -134.26, 'y': 0.2563}
        1 {'x': -12.0, 'z': 0.3}
        -1 {'y': 0.25}

    References:
        1. [LIBSVM documentation](https://www.csie.ntu.edu.tw/~cjlin/libsvm/)

    """

    # If a file is not opened, then we open it
    buffer = filepath_or_buffer
    if not hasattr(buffer, 'read'):
        buffer = utils.open_filepath(buffer, compression)

    def split_pair(pair):
        name, value = pair.split(':')
        value = float(value)
        return name, value

    for line in buffer:

        # Remove carriage return and whitespace
        line = line.rstrip()
        # Remove potential end of line comments
        line = line.split('#')[0]

        y, x = line.split(' ', maxsplit=1)
        y = target_type(y)
        x = dict([split_pair(pair) for pair in x.split(' ')])
        yield x, y

    # Close the file if we opened it
    if buffer is not filepath_or_buffer:
        buffer.close()



import collections
import random

import mmh3

from .. import base


__all__ = ['FeatureHasher']


class FeatureHasher(base.Transformer):
    """Implements the hashing trick.

    Each pair of (name, value) features is hashed into a random integer. A module operator is then
    used to make sure the hash is in a certain range. We use the Murmurhash implementation from
    scikit-learn.

    Parameters:
        n_features: The number by which each hash will be moduloed by.
        alternate_sign: Whether or not half of the hashes will be negated.
        seed: Set the seed to produce identical results.

    Example:

        >>> import creme

        >>> hasher = creme.preprocessing.FeatureHasher(n_features=10, seed=0)

        >>> X = [
        ...     {'dog': 1, 'cat': 2, 'elephant': 4},
        ...     {'dog': 2, 'run': 5}
        ... ]
        >>> for x in X:
        ...     print(hasher.transform_one(x))
        defaultdict(<class 'int'>, {0: 1, 3: -2, 4: 4})
        defaultdict(<class 'int'>, {0: -3})

    References:
        1. [Wikipedia article on feature vectorization using the hashing trick](https://www.wikiwand.com/en/Feature_hashing#/Feature_vectorization_using_hashing_trick)

    """

    def __init__(self, n_features=1048576, alternate_sign=True, seed: int = None):
        self.n_features = n_features
        self.alternate_sign = alternate_sign
        self.seed = seed
        self._hash_seed = random.Random(seed).randint(0, 2 ** 10)

    def transform_one(self, x):

        x_hashed = collections.defaultdict(int)

        for feature, value in x.items():

            if isinstance(value, str):
                feature = f'{feature}={value}'
                value = 1

            if value == 0:
                continue

            h = mmh3.hash(feature, seed=self._hash_seed)
            i = h % self.n_features

            if self.alternate_sign:
                value *= (h >= 0) * 2 - 1

            x_hashed[i] += value

        return x_hashed



import random

from .. import base


class SEA(base.SyntheticDataset):
    """SEA synthetic dataset.

    Each observation is composed of 3 features. Only the first two features are relevant. The
    target is binary, and is positive if the sum of the features exceeds a certain threshold. There
    are 4 thresholds to choose from. Concept drift can be introduced by switching the threshold
    anytime during the stream.

    Parameters:
        variant: Determines the classification function to use. Possible choices are 0, 1, 2, 3.
        noise: Determines the amount of observations for which the target sign will be flipped.
        seed: Random seed number used for reproducibility.

    Example:

        >>> from creme import datasets

        >>> dataset = datasets.synth.SEA(variant=0, seed=42)

        >>> for x, y in dataset.take(5):
        ...     print(x, y)
        {0: 6.39426, 1: 0.25010, 2: 2.75029} False
        {0: 2.23210, 1: 7.36471, 2: 6.76699} True
        {0: 8.92179, 1: 0.86938, 2: 4.21921} True
        {0: 0.29797, 1: 2.18637, 2: 5.05355} False
        {0: 0.26535, 1: 1.98837, 2: 6.49884} False

    References:
        1. [A Streaming Ensemble Algorithm (SEA) for Large-Scale Classification](http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.482.3991&rep=rep1&type=pdf)

    """

    def __init__(self, variant: int, noise=0., seed: int = None):

        super().__init__(n_features=3, task=base.BINARY_CLF)

        if variant not in (0, 1, 2, 3):
            raise ValueError('Unknown variant, possible choices are: 0, 1, 2, 3')

        self.variant = variant
        self.noise = noise
        self.seed = seed
        self._threshold = {0: 8, 1: 9, 2: 7, 3: 9.5}[variant]
        self._rng = random.Random(seed)

    def __iter__(self):

        while True:

            x = {i: self._rng.uniform(0, 10) for i in range(3)}
            y = x[0] + x[1] > self._threshold

            if self.noise and self._rng.random() < self.noise:
                y = not y

            yield x, y

    @property
    def _repr_title(self):
        return f'{self.__class__.__name__} dataset, variant #{self.variant}'



from creme import stream

from . import base


class HTTP(base.RemoteDataset):
    """HTTP dataset of the KDD 1999 cup.

    The goal is to predict whether or not an HTTP connection is anomalous or not. The dataset only
    contains 2,211 (0.4%) positive labels.

    References:
        1. [HTTP (KDDCUP99) dataset](http://odds.cs.stonybrook.edu/http-kddcup99-dataset/)

    """

    def __init__(self):
        super().__init__(
            n_samples=567_498,
            n_features=3,
            task=base.BINARY_CLF,
            url='https://maxhalford.github.io/files/datasets/kdd99_http.zip',
            size=32400738,
            filename='kdd99_http.csv'
        )

    def _iter(self):
        converters = {'duration': float, 'src_bytes': float, 'dst_bytes': float, 'service': int}
        return stream.iter_csv(self.path, target='service', converters=converters)



from . import base


__all__ = ['Accuracy']


class Accuracy(base.MeanMetric, base.MultiClassMetric):
    """Accuracy score, which is the percentage of exact matches.

    Example:

        >>> from creme import metrics

        >>> y_true = [True, False, True, True, True]
        >>> y_pred = [True, True, False, True, True]

        >>> metric = metrics.Accuracy()
        >>> for yt, yp in zip(y_true, y_pred):
        ...     metric = metric.update(yt, yp)

        >>> metric
        Accuracy: 60.00%

    """

    _fmt = '.2%'  # will output a percentage, e.g. 0.427 will become "42,7%"

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def _eval(self, y_true, y_pred):
        return y_true == y_pred



import surprise

from . import base


class SurpriseWrapper(surprise.AlgoBase):
    """A wrapper to provide compatibility with surprise.

    Parameters:
        creme_recommender

    """

    def __init__(self, creme_recommender: base.Recommender):
        self.creme_recommender = creme_recommender

    def fit(self, trainset):
        surprise.AlgoBase.fit(self, trainset)
        for user, item, y in trainset.all_ratings():
            self.creme_recommender.fit_one({'user': user, 'item': item}, y)
        return self

    def estimate(self, user, item):
        return self.creme_recommender.predict_one({'user': user, 'item': item})



import abc
import copy
import inspect
import sys
import types
import typing

from .. import utils


DEFAULT_TAGS = {
    'handles_text': False,
    'requires_positive_data': False,
    'handles_categorical_features': False
}


class Estimator(abc.ABC):
    """An estimator."""

    def _is_supervised(self):
        """Indicates whether or not the estimator is supervised or not.

        This is useful internally for determining if an estimator expects to be provided with a `y`
        value in it's `fit_one` method. For instance we use this in a pipeline to know whether or
        not we should pass `y` to an estimator or not.

        """
        return True

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return _repr_obj(obj=self, params=self._get_params())

    def __or__(self, other):
        """Merges with another Transformer into a Pipeline."""
        from .. import compose
        if isinstance(other, compose.Pipeline):
            return other.__ror__(self)
        return compose.Pipeline(self, other)

    def __ror__(self, other):
        """Merges with another Transformer into a Pipeline."""
        from .. import compose
        if isinstance(other, compose.Pipeline):
            return other.__or__(self)
        return compose.Pipeline(other, self)

    def _get_params(self) -> typing.Dict[str, typing.Any]:
        return {
            name: getattr(self, name)
            for name, param in inspect.signature(self.__init__).parameters.items()  # type: ignore
            if param.kind != param.VAR_KEYWORD
        }

    def _set_params(self, new_params: typing.Optional[typing.Dict[str, typing.Any]] = None) -> 'Estimator':
        """Returns a new instance with the current parameters as well as new ones.

        The algorithm will be recursively called down `Pipeline`s and `TransformerUnion`s.

        Example:

            >>> from creme import linear_model
            >>> from creme import optim
            >>> from creme import preprocessing

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LinearRegression(
            ...         optimizer=optim.SGD(lr=0.042),
            ...     )
            ... )

            >>> new_params = {
            ...     'LinearRegression': {
            ...         'l2': .001
            ...     }
            ... }

            >>> model._set_params(new_params)
            Pipeline (
              StandardScaler (),
              LinearRegression (
                optimizer=SGD (
                  lr=Constant (
                    learning_rate=0.042
                  )
                )
                loss=Squared ()
                l2=0.001
                intercept=0.
                intercept_lr=Constant (
                  learning_rate=0.01
                )
                clip_gradient=1e+12
                initializer=Zeros ()
              )
            )

        """

        if new_params is None:
            new_params = {}

        params = {
            **self._get_params(),
            **new_params
        }

        return self.__class__(**copy.deepcopy(params))  # type: ignore

    @property
    def _tags(self) -> typing.Dict[str, bool]:
        """Returns the estimator's tags.

        Tags can be used to specify what kind of inputs an estimator is able to process. For
        instance, some estimators can handle text, whilst others require positive numeric data.
        Inheriting from `base.Estimator` will imply a set of default tags which can be overriden
        by implementing the `base.Estimator._more_tags` method.

        """

        tags: typing.Dict[str, bool] = {}

        for base_class in inspect.getmro(self.__class__)[1:]:
            if isinstance(base_class, Estimator):
                _update_if_consistent(tags, base_class._more_tags)

        _update_if_consistent(tags, self._more_tags)

        return {**DEFAULT_TAGS, **tags}

    @property
    def _more_tags(self) -> typing.Dict[str, bool]:
        """Specific tags for this estimator."""
        return {}

    @property
    def _memory_usage(self) -> str:
        """Returns the memory usage in a human readable format."""

        def get_size(obj, seen=None):
            """Recursively finds size of objects"""
            size = sys.getsizeof(obj)
            if seen is None:
                seen = set()
            obj_id = id(obj)
            if obj_id in seen:
                return 0
            # Important mark as seen *before* entering recursion to gracefully handle
            # self-referential objects
            seen.add(obj_id)
            if isinstance(obj, dict):
                size += sum([get_size(v, seen) for v in obj.values()])
                size += sum([get_size(k, seen) for k in obj.keys()])
            elif hasattr(obj, '__dict__'):
                size += get_size(vars(obj), seen)
            elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
                size += sum([get_size(i, seen) for i in obj])
            return size

        mem_usage = get_size(self)
        return utils.pretty.humanize_bytes(mem_usage)


def _update_if_consistent(dict1: dict, dict2: dict):
    common_keys = set(dict1.keys()).intersection(dict2.keys())
    for key in common_keys:
        if dict1[key] != dict2[key]:
            raise TypeError(
                f'Inconsistent values for tag {key}: {dict1[key]} != {dict2[key]}')
    dict1.update(dict2)


def _repr_obj(obj, params=None, show_modules: bool = False, depth: int = 0) -> str:
    """Returns a pretty representation of an object."""

    rep = f'{obj.__class__.__name__} ('
    if show_modules:
        rep = f'{obj.__class__.__module__}.{rep}'
    tab = '\t'

    if params is None:
        params = {
            name: getattr(obj, name)
            for name, param in inspect.signature(obj.__init__).parameters.items()  # type: ignore
            if not (
                param.name == 'args' and param.kind == param.VAR_POSITIONAL or
                param.name == 'kwargs' and param.kind == param.VAR_KEYWORD
            )
        }

    n_params = 0

    for name, val in params.items():

        n_params += 1

        # Prettify the attribute when applicable
        if isinstance(val, types.FunctionType):
            val = val.__name__
        if isinstance(val, str):
            val = f'"{val}"'
        elif isinstance(val, float):
            val = (
                f'{val:.0e}'
                if (val > 1e5 or (val < 1e-4 and val > 0)) else
                f'{val:.6f}'.rstrip('0')
            )
        elif isinstance(val, set):
            val = sorted(val)
        elif hasattr(val, '__class__') and 'creme.' in str(type(val)):
            val = _repr_obj(obj=val, show_modules=show_modules, depth=depth + 1)

        rep += f'\n{tab * (depth + 1)}{name}={val}'

    if n_params:
        rep += f'\n{tab * depth}'
    rep += ')'

    return rep.expandtabs(2)



import collections
import typing

from . import base
from . import pearson


class AutoCorr(base.Univariate):
    """Measures the serial correlation.

    This method computes the Pearson correlation between the current value and the value seen `n`
    steps before.

    Parameters:
        lag

    Example:

        The following examples are taken from the [pandas documentation](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.autocorr.html).

        >>> from creme import stats

        >>> auto_corr = stats.AutoCorr(lag=1)
        >>> for x in [0.25, 0.5, 0.2, -0.05]:
        ...     print(auto_corr.update(x).get())
        0
        0
        -1.0
        0.103552

        >>> auto_corr = stats.AutoCorr(lag=2)
        >>> for x in [0.25, 0.5, 0.2, -0.05]:
        ...     print(auto_corr.update(x).get())
        0
        0
        0
        -1.0

        >>> auto_corr = stats.AutoCorr(lag=1)
        >>> for x in [1, 0, 0, 0]:
        ...     print(auto_corr.update(x).get())
        0
        0
        0
        0

    """

    def __init__(self, lag: int):
        self.window: typing.Deque[float] = collections.deque(maxlen=lag)
        self.lag = lag
        self.pearson = pearson.PearsonCorr(ddof=1)

    @property
    def name(self):
        return f'autocorr_{self.lag}'

    def update(self, x):

        # The correlation can be update once enough elements have been seen
        if len(self.window) == self.lag:
            self.pearson.update(x, self.window[0])

        # Add x to the window
        self.window.append(x)

        return self

    def get(self):
        return self.pearson.get()



"""Utilities for unit testing and sanity checking estimators."""
import copy
import functools
import inspect
import itertools
import math
import pickle


__all__ = [
    'check_estimator',
    'guess_model'
]


def guess_model(model):
    """Extracts the relevant part model.

    Sometimes we need to check if a model can perform regression, classification, etc. When the
    model is a pipeline, then this can be checked by looking at the final step of the pipeline.

    This function is highly succeptible to be modified or disappear altogether. The only reason it
    exists is that we can't do `isinstance(pipeline, base.Regressor)`.

    Parameters:
        model

    """

    from creme import base
    from creme import compose

    if isinstance(model, (base.MultiOutputRegressor, base.MultiOutputClassifier)):
        return model
    elif isinstance(model, compose.Pipeline):
        return guess_model(list(model.steps.values())[-1])
    elif isinstance(model, base.Wrapper):
        return guess_model(model._wrapped_model)
    return model


def yield_datasets(model):

    from creme import base
    from creme import datasets
    from creme import stream
    from sklearn import datasets as sk_datasets

    model = guess_model(model)

    if isinstance(model, (base.BinaryClassifier, base.MultiClassifier)):
        yield datasets.Phishing()
    if isinstance(model, base.MultiClassifier):
        yield datasets.ImageSegments().take(500)
    if isinstance(model, base.Regressor):
        yield datasets.TrumpApproval()
    if isinstance(model, base.MultiOutputRegressor):
        yield stream.iter_sklearn_dataset(sk_datasets.load_linnerud())
    if isinstance(model, base.MultiOutputClassifier):
        yield datasets.Music()


def check_fit_one(model, dataset):

    klass = model.__class__

    for x, y in dataset:

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        model = model.fit_one(x, y)

        # Check the model returns itself
        assert isinstance(model, klass)

        # Check fit_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_predict_proba_one(classifier, dataset):

    for x, y in dataset:

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        classifier = classifier.fit_one(x, y)
        y_pred = classifier.predict_proba_one(x)

        # Check the probabilities are coherent
        assert isinstance(y_pred, dict)
        assert math.isclose(sum(y_pred.values()), 1.)
        for proba in y_pred.values():
            assert 0. <= proba <= 1.

        # Check predict_proba_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_predict_proba_one_binary(classifier, dataset):

    for x, y in dataset:
        y_pred = classifier.predict_proba_one(x)
        classifier = classifier.fit_one(x, y)
        assert len(y_pred) == 2
        assert True in y_pred
        assert False in y_pred


def check_debug_one(model, dataset):
    for x, y in dataset:
        model.debug_one(x)
        model.fit_one(x, y)
        model.debug_one(x)
        break


def check_pickling(model, dataset):
    assert isinstance(pickle.loads(pickle.dumps(model)), model.__class__)
    for x, y in dataset:
        model.predict_one(x)
        model.fit_one(x, y)
    assert isinstance(pickle.loads(pickle.dumps(model)), model.__class__)


def check_repr_works(model):
    rep = repr(model)
    assert isinstance(rep, str)


def check_str_works(model):
    assert isinstance(str(model), str)


def check_tags(model):
    """Checks that the `_tags` property works."""
    assert isinstance(model._tags, dict)


def check_set_params_idempotent(model):
    assert len(model.__dict__) == len(model._set_params().__dict__)


def yield_checks(model):
    """Generates unit tests that can be applied to a given model.

    Parameters:
        model (base.Estimator)

    """

    from creme import base

    yield check_repr_works
    yield check_str_works
    yield check_tags
    yield check_set_params_idempotent

    for dataset in yield_datasets(model):

        def with_dataset(method):
            check = functools.partial(check_fit_one, dataset=dataset)
            functools.update_wrapper(check, method)
            return check

        yield with_dataset(check_fit_one)
        yield with_dataset(check_pickling)
        if hasattr(model, 'debug_one'):
            yield with_dataset(check_debug_one)

        model = guess_model(model)

        if isinstance(model, base.Classifier):
            yield with_dataset(check_predict_proba_one)

        if not isinstance(model, base.MultiClassifier):  # only binary classifiers
            yield with_dataset(check_predict_proba_one_binary)


def check_estimator(model):
    """Check if a model adheres to `creme`'s conventions.

    Parameters:
        model

    """
    for check in yield_checks(model):
        check(copy.deepcopy(model))



import collections
import copy
import statistics

import numpy as np

from .. import base


__all__ = ['BaggingClassifier', 'BaggingRegressor']


class BaseBagging(base.Wrapper, base.Ensemble):

    def __init__(self, model, n_models=10, seed=None):
        super().__init__(copy.deepcopy(model) for _ in range(n_models))
        self.n_models = n_models
        self.model = model
        self.seed = seed
        self._rng = np.random.RandomState(seed)

    @property
    def _wrapped_model(self):
        return self.model

    def fit_one(self, x, y):

        for model in self:
            for _ in range(self._rng.poisson(1)):
                model.fit_one(x, y)

        return self


class BaggingClassifier(BaseBagging, base.Classifier):
    """Online bootstrap aggregation for classification.

    For each incoming observation, each model's `fit_one` method is called `k` times where
    `k` is sampled from a Poisson distribution of parameter 1. `k` thus has a 36% chance of
    being equal to 0, a 36% chance of being equal to 1, an 18% chance of being equal to 2, a 6%
    chance of being equal to 3, a 1% chance of being equal to 4, etc. You can do
    `scipy.stats.poisson(1).pmf(k)` to obtain more detailed values.

    Parameters:
        model: The classifier to bag.
        n_models: The number of models in the ensemble.
        seed: Random number generator seed for reproducibility.

    Example:

        In the following example three logistic regressions are bagged together. The performance is
        slightly better than when using a single logistic regression.

        >>> from creme import datasets
        >>> from creme import ensemble
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> model = ensemble.BaggingClassifier(
        ...     model=(
        ...         preprocessing.StandardScaler() |
        ...         linear_model.LogisticRegression()
        ...     ),
        ...     n_models=3,
        ...     seed=42
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.877788

        >>> print(model)
        BaggingClassifier(StandardScaler | LogisticRegression)

    References:
        1. [Oza, N.C., 2005, October. Online bagging and boosting. In 2005 IEEE international conference on systems, man and cybernetics (Vol. 3, pp. 2340-2345). Ieee.](https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf)

    """

    def __init__(self, model: base.Classifier, n_models=10, seed: int = None):
        super().__init__(model, n_models, seed)

    def predict_proba_one(self, x):
        """Averages the predictions of each classifier."""

        y_pred = collections.Counter()
        for classifier in self:
            y_pred.update(classifier.predict_proba_one(x))

        total = sum(y_pred.values())
        if total > 0:
            return {label: proba / total for label, proba in y_pred.items()}
        return y_pred


class BaggingRegressor(BaseBagging, base.Regressor):
    """Online bootstrap aggregation for regression.

    For each incoming observation, each model's `fit_one` method is called `k` times where
    `k` is sampled from a Poisson distribution of parameter 1. `k` thus has a 36% chance of
    being equal to 0, a 36% chance of being equal to 1, an 18% chance of being equal to 2, a 6%
    chance of being equal to 3, a 1% chance of being equal to 4, etc. You can do
    `scipy.stats.poisson(1).pmf(k)` for more detailed values.

    Parameters:
        model: The regressor to bag.
        n_models: The number of models in the ensemble.
        seed: Random number generator seed for reproducibility.

    Example:

        In the following example three logistic regressions are bagged together. The performance is
        slightly better than when using a single logistic regression.

        >>> from creme import datasets
        >>> from creme import ensemble
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.TrumpApproval()
        >>> model = preprocessing.StandardScaler()
        >>> model |= ensemble.BaggingRegressor(
        ...     model=linear_model.LinearRegression(intercept_lr=0.1),
        ...     n_models=3,
        ...     seed=42
        ... )
        >>> metric = metrics.MAE()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MAE: 0.641799

    References:
        1. [Oza, N.C., 2005, October. Online bagging and boosting. In 2005 IEEE international conference on systems, man and cybernetics (Vol. 3, pp. 2340-2345). Ieee.](https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf)

    """

    def __init__(self, model: base.Regressor, n_models=10, seed: int = None):
        super().__init__(model, n_models, seed)

    def predict_one(self, x):
        """Averages the predictions of each regressor."""
        return statistics.mean((regressor.predict_one(x) for regressor in self))
